/*
 * generated by Xtext
 */
package fr.inria.phoenix.diasuite.diaswarm.generator

import com.google.inject.Inject
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.AbstractElement
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.ContextDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.ControllerDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DataTypeRef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DeviceType
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DomainModel
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.EnumDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.InteractionContractDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.PushInteractionContract
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.StructDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.StructFieldDef
import fr.inria.phoenix.diasuite.diaswarm.generator.template.IConstant
import fr.inria.phoenix.diasuite.diaswarm.generator.template.ISetGenerator
import fr.inria.phoenix.diasuite.diaswarm.generator.template.context.DiaSwarmContext
import fr.inria.phoenix.diasuite.diaswarm.generator.template.controller.DiaSwarmController
import fr.inria.phoenix.diasuite.diaswarm.generator.template.datatype.DiaSwarmDatatype
import fr.inria.phoenix.diasuite.diaswarm.generator.template.deployment.DiaSwarmDeployment
import fr.inria.phoenix.diasuite.diaswarm.generator.template.device.DiaSwarmDevice
import java.util.ArrayList
import java.util.HashMap
import java.util.List
import java.util.Map
import org.eclipse.emf.ecore.resource.Resource
import org.eclipse.emf.ecore.resource.ResourceSet
import org.eclipse.xtext.generator.IFileSystemAccess

import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.context.DiaSwarmContextHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.controller.DiaSwarmControllerHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.datatype.DiaSwarmDatatypeHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.deployment.DiaSwarmDeploymentHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.device.DiaSwarmDeviceHelper.*

/**
 * Generates code from your model files on save.
 * 
 * see http://www.eclipse.org/Xtext/documentation.html#TutorialCodeGeneration
 */
class DiaSwarmGenerator implements ISetGenerator {

	@Inject extension DiaSwarmDatatype
	@Inject extension DiaSwarmDevice
	@Inject extension DiaSwarmContext
	@Inject extension DiaSwarmController
	@Inject extension DiaSwarmDeployment
//	@Inject extension MockDeviceGen
//	@Inject extension MockClassGen
	public static boolean inEclipse = true;

	override void doGenerate(Resource resource, IFileSystemAccess fsa) {
		println("Entered the doGenerate Resource method !")
		println("[Resource] Model iterable size : " + resource.allContents.toIterable.filter(typeof(DomainModel)).size)

		val model = resource.allContents.toIterable.filter(typeof(DomainModel)).findFirst[model|model != null]

		// Generating code for DiaSwarm components and data types.
		for (entity : resource.allContents.toIterable.filter(typeof(AbstractElement)))
			entity.generateCode(fsa)

		// Generating deployment code for a DiaSwarm application.
		model.generateCode(fsa)

	// Generating code for mock devices.
//		fsa.generateFile(generateMockFileName, model.generateMockClass)
	}

	override void doGenerate(ResourceSet input, IFileSystemAccess fsa) {
		println("Entered the doGenerate ResourceSet method !")

		// Generating code for DiaSwarm components and data types.
		for (entity : input.allContents.toIterable.filter(typeof(AbstractElement)))
			entity.generateCode(fsa)

		// Generating deployment code for a DiaSwarm application.
		input.allContents.toIterable.filter(typeof(DomainModel)).generateCode(fsa)

		// Generating code for mock devices.
		// println("[ResourceSet] Model iterable size : " + models.size)
		val List<DomainModel> modelList = new ArrayList<DomainModel>();
		for (model : input.allContents.toIterable.filter(typeof(DomainModel)))
			modelList.add(model);

//		fsa.generateFile(generateMockFileName, modelList.generateMockClass)
	}

	/**
	 * Generates code for DiaSwarm structures.
	 */
	def dispatch generateCode(StructDef structure, IFileSystemAccess fsa) {
		fsa.generateFile(structure.generateFileName, structure.generateStructure)
	}

	/**
	 * Generates code for DiaSwarm enumerations.
	 */
	def dispatch generateCode(EnumDef enumeration, IFileSystemAccess fsa) {
		fsa.generateFile(enumeration.generateFileName, enumeration.generateEnumeration)
	}

	/**
	 * Generates code for DiaSwarm devices.
	 */
	def dispatch generateCode(DeviceType device, IFileSystemAccess fsa) {
		fsa.generateFile(device.generateFileName, device.generateDevice)

		// Here we generate individual classes for every device source
		device.allSources.forEach([ source |
			if (!source.indices.empty)
				fsa.generateFile(source.generateFileName(device, "Indices"), source.generateDeviceSourceIndices(device))
			fsa.generateFile(source.generateFileName(device, "From" + device.name),
				source.generateDeviceSourceValue(device))
		])

//		fsa.generateFile(device.generateMockDeviceFileName, device.generateMock)
	}

	/**
	 * Generates code for DiaSwarm contexts.
	 */
	def dispatch generateCode(ContextDef context, IFileSystemAccess fsa) {
		fsa.generateFile(context.generateFileName, context.generateContext)

		if (!context.getIndices().isEmpty()) {
			fsa.generateFile(context.generateFileNameForContextFeature("Indices"), context.contextIndices)
		}
		fsa.generateFile(context.generateFileNameForContextFeature("Value"), context.contextValue)

		val mapReduceBehavior = context.behaviors.filter [
			((interaction.equals(IConstant.PERIODIC_INTERACTION) ||
				interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION)) &&
				((contract as PushInteractionContract).providedData.map ||
					(contract as PushInteractionContract).providedData.reduce))
			];

			if (!mapReduceBehavior.isEmpty) {
				fsa.generateFile(context.generateFileNameForContextFeature("Job"),
					context.generateJob(mapReduceBehavior.last.contract as PushInteractionContract))
				fsa.generateFile(context.generateFileNameForContextFeature("Adapter"),
					context.generateAdapter(mapReduceBehavior.last.contract as PushInteractionContract))
				fsa.generateFile(context.generateFileNameForContextFeature("ResultsParser"),
					context.generateResultsParser(mapReduceBehavior.last.contract as PushInteractionContract))
				fsa.generateFile(context.generateFileNameForContextFeature("OutputFormat"),
					context.generateOutputFormat(mapReduceBehavior.last.contract as PushInteractionContract))
				fsa.generateFile(context.generateFileNameForContextFeature("RecordWriter"),
					context.generateRecordWriter(mapReduceBehavior.last.contract as PushInteractionContract))
				if ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.mapOutValueType.type !=
					null) {
//				// if it is an enumeration
//				if ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.mapOutValueType.
//					type instanceof EnumDef) {
//					val enumeration = ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.
//						mapOutValueType.type as EnumDef);
//					fsa.generateFile(enumeration.generateFileName("Writable"), enumeration.generateWritableEnumeration)
//				}
					// it is a source
					if ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.mapOutKeyType !=
						null &&
						(mapReduceBehavior.last.contract as PushInteractionContract).providedData.mapOutKeyType.
							type instanceof StructDef) {
						val structure = ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.
							mapOutKeyType.type as StructDef)
						fsa.generateFile(structure.generateFileName("Writable"), structure.generateWritableStructure)

						for (StructFieldDef field : ((mapReduceBehavior.last.contract as PushInteractionContract).
							providedData.mapOutKeyType.type as StructDef).structFields) {
							if (field.dataTypeRef.type != null && field.dataTypeRef.type instanceof StructDef) {
								val fieldStructure = (field.dataTypeRef.type as StructDef)
								fsa.generateFile(fieldStructure.generateFileName("Writable"),
									fieldStructure.generateWritableStructure)
							}
						}
					}

					if ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.mapOutValueType.
						type instanceof StructDef) {
						val structure = ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.
							mapOutValueType.type as StructDef)
						fsa.generateFile(structure.generateFileName("Writable"), structure.generateWritableStructure)

						for (StructFieldDef field : ((mapReduceBehavior.last.contract as PushInteractionContract).
							providedData.mapOutValueType.type as StructDef).structFields) {
							if (field.dataTypeRef.type != null && field.dataTypeRef.type instanceof StructDef) {
								val fieldStructure = (field.dataTypeRef.type as StructDef)
								fsa.generateFile(fieldStructure.generateFileName("Writable"),
									fieldStructure.generateWritableStructure)
							}
						}
					}

					if ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.reduceType.
						type instanceof StructDef) {
						val structure = ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.
							reduceType.type as StructDef)
						fsa.generateFile(structure.generateFileName("Writable"), structure.generateWritableStructure)

						for (StructFieldDef field : ((mapReduceBehavior.last.contract as PushInteractionContract).
							providedData.reduceType.type as StructDef).structFields) {
							if (field.dataTypeRef.type != null && field.dataTypeRef.type instanceof StructDef) {
								val fieldStructure = (field.dataTypeRef.type as StructDef)
								fsa.generateFile(fieldStructure.generateFileName("Writable"),
									fieldStructure.generateWritableStructure)
							}
						}
					}
				}
			}
		}

		/**
		 * Generates code for DiaSwarm controllers.
		 */
		def dispatch generateCode(ControllerDef controller, IFileSystemAccess fsa) {
			fsa.generateFile(controller.generateFileName, controller.generateController)
		}

		/**
		 * Generates application deployment code.
		 */
		def dispatch generateCode(Iterable<DomainModel> models, IFileSystemAccess fsa) {
			var deviceSourceBufferSize = new HashMap<String, Integer>();
			var deviceSourceToGroupingAttributes = new HashMap<String, Map<String, String>>();
			var deviceSourceType = new HashMap<String, DataTypeRef>();
			var List<DomainModel> modelList = new ArrayList<DomainModel>();

			for (DomainModel model : models) {
				modelList.add(model);
			}

			for (ContextDef context : modelList.contexts) {
				for (InteractionContractDef behavior : context.behaviors) {
					if ((behavior.interaction.equals(IConstant.PERIODIC_INTERACTION) ||
						(behavior.interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION) &&
							(behavior.contract as PushInteractionContract).providedData.periodicity != null))) {
						var requirement = (behavior.contract as PushInteractionContract).providedData
						var sourceRef = requirement.dataSource.sourceRef

						if (sourceRef != null) {
							var deviceName = sourceRef.device.name
							var sourceName = sourceRef.source.name
							var sourceType = sourceRef.source.dataTypeRef

							deviceSourceType.put(sourceName + "BufferFor" + deviceName, sourceType)

							println("sourceName : " + sourceName);
							var devicePublishingPeriodicity = requirement.dataSource.periodicity
							var groupingPeriodicity = requirement.periodicity
							var bufferSize = 1

							if (devicePublishingPeriodicity != null && groupingPeriodicity != null)
								bufferSize = groupingPeriodicity.periodicityInSeconds /
									devicePublishingPeriodicity.periodicityInSeconds
							else if(groupingPeriodicity != null)
								bufferSize = -1;

							var lastBufferSize = deviceSourceBufferSize.get(sourceName + "BufferFor" + deviceName)

							if (lastBufferSize == null || (lastBufferSize < bufferSize || bufferSize == -1))
								deviceSourceBufferSize.put(sourceName + "BufferFor" + deviceName, bufferSize)

							var groupingAttribute = requirement.groupingAttribute

							if (groupingAttribute != null) {
								var groupingAttributeName = groupingAttribute.name;
								var String groupingAttributeType;
								var attributes = deviceSourceToGroupingAttributes.get(sourceName + "BufferFor" +
									deviceName)

								if (attributes == null) {
									attributes = new HashMap<String, String>();
									deviceSourceToGroupingAttributes.put(sourceName + "BufferFor" + deviceName,
										attributes)
								}

								if (groupingAttribute.attrTypeRef.type != null)
									groupingAttributeType = groupingAttribute.attrTypeRef.fullName
								else
									groupingAttributeType = groupingAttribute.attrTypeRef.primitiveTypeRef.literal

								attributes.put(groupingAttributeName, groupingAttributeType)
							}
						}
					}
				}
			}

			fsa.generateFile(modelList.generateDeploymentFileName,
				modelList.generateDeploymentCode(deviceSourceBufferSize.keySet))

			for (bufferName : deviceSourceBufferSize.keySet) {
				var bufferSize = deviceSourceBufferSize.get(bufferName)
				var attributes = deviceSourceToGroupingAttributes.get(bufferName)
				var sourceType = deviceSourceType.get(bufferName)

				println("Buffer name : " + bufferName)
				println("Buffer size : " + bufferSize)

				println("---------------------------------------------")

				fsa.generateFile(modelList.generateBufferFileName(bufferName),
					modelList.generateBufferForDeviceSource(bufferName, (bufferName.split("BufferFor")).get(0),
						sourceType, bufferSize, attributes))
			}
		}
	}
	