package fr.inria.phoenix.diasuite.diaswarm.generator.template

import fr.inria.phoenix.diasuite.diaswarm.diaswarm.ActionDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.AttributeDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.ContextDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DataTypeRef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DeviceType
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.SourceDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.VariableDef
import java.util.Collection
import java.util.List
import org.eclipse.emf.ecore.EObject

import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.datatype.DiaSwarmDatatypeHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.device.DiaSwarmDeviceHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.Utils.*



// Generation of classes that are common to devices, context and controllers
public class GenericTemplate {
	
	/* Create the class container for indices */
    def static String classForIndices(List<VariableDef> indices, 
        String packageName, String objectName, EObject object, 
        String comment) 
	'''
        package «packageName»;
        
        import java.io.Serializable;
        
«««        «object.javadoc(comment)»
        public final class «objectName.toFirstUpper»Indices implements Serializable {
            private static final long serialVersionUID = 0;
            
            public «objectName.toFirstUpper»Indices(«indices.parameters») {
                «FOR idx : indices»
                    this._«idx.name» = «idx.name»;
                «ENDFOR»
            }
            
            «FOR idx : indices»
                // Code for index «idx.name»
                private «idx.dataTypeRef.fullName» _«idx.name»;
                
«««                /**
«««                 * Get the <code>«idx.name»</code> index
«««                 «IF !Utils::getComments(idx, 1).nullOrEmpty»
«««                 * 
«««                 * <p>
«««                 * «Utils::getComments(idx, 1).replace("\n", "\n* ")»
«««                 «ENDIF»
«««                 * 
«««                 * @return the value of the <code>«idx.name»</code> index
«««                 */
                public «idx.dataTypeRef.fullName» «idx.name»() {
                    return _«idx.name»;
                }
                
«««                /**
«««                 * Set the <code>«idx.name»</code> index
«««                 «IF !Utils::getComments(idx, 1).nullOrEmpty»
«««                 * 
«««                 * <p>
«««                 * «Utils::getComments(idx, 1).replace("\n", "\n* ")»
«««                 «ENDIF»
«««                 * 
«««                 * @param new«idx.name.toFirstUpper» the new value for the <code>«idx.name»</code> index
«««                 */
                public void set«idx.name.toFirstUpper»(«idx.dataTypeRef.fullName» new«idx.name.toFirstUpper») {
                    this._«idx.name» = new«idx.name.toFirstUpper»;
                }
                // End of code for index «idx.name»
                
            «ENDFOR»
«««            «equalsAndHashCode(objectName.toFirstUpper + "Indices", indices.map["_" + name])»
        }
    '''
    
    
	/* Generate the container class for publishing a value */
	def static valueClass(String name, String className, String packageName, String description, EObject obj,
		String comment, DataTypeRef valueType, Collection<VariableDef> indices, String extraCons, String extraArg,
		String misc, String imports, String extraComments) {
		val hasIndices = (indices != null && indices.size > 0);

		'''
			package «packageName»;
			
			import java.io.Serializable;
			«imports»
«««			/**
«««			 * «comment.replaceAll("\n", "\n* ")»
«««			 *
«««			 * <pre>
«««			 * «obj.stripComments.replaceAll("\n", "\n* ")»
«««			 * </pre>
«««			 */
			
			public final class «className» implements Serializable {
			    private static final long serialVersionUID = 0;
			    
			    private «valueType.fullName» value;
			    
			    /**
			     * Get the value of «description»
			     «IF !extraComments.nullOrEmpty»
			     	* 
			     	* <p>
			     	* «extraComments.split("\n").join("\n* ")»
			     «ENDIF»
			     * 
			     * @return the value of «description»
			     */
			    public «valueType.fullName» value() {
			        return value;
			    }
			    
			    «IF hasIndices»
			    	private «name.toFirstUpper»Indices indices;
			    	
			    	/**
			    	 * Get the value of the indices of «description»
			    	 * 
			    	 * @return the value of the indices
			    	 */
			    	public «name.toFirstUpper»Indices indices() {
			    	    return indices;
			    	}
			    	
			    	public «className»(«extraArg»«valueType.fullName» value, «name.toFirstUpper»Indices indices) {
			    	    «extraCons»this.value = value;
			    	    this.indices = indices;
			    	}
			    	
			    	public «className»(«extraArg»«valueType.fullName» value,
			    	        «indices.map[x | x.dataTypeRef.fullName + " " + x.name].join(",\n")») {
			    	«extraCons»this.value = value;
			    	this.indices = new «name.toFirstUpper»Indices(«indices.variableNames»);
			    	}
			    «ELSE»
			    	public «className»(«extraArg»«valueType.fullName» value) {
			    	    «extraCons»this.value = value;
			    	}
			    «ENDIF»
			    
			    @Override
			    public int hashCode() {
			        final int prime = 31;
			        int result = 1;
			        result = prime * result + ((value == null) ? 0 : value.hashCode());
			        «IF hasIndices»
			        	result = prime * result + ((indices == null) ? 0 : indices.hashCode());
			        «ENDIF»
			        return result;
			    }
			    
			    @Override
			    public boolean equals(Object obj) {
			        if (this == obj)
			            return true;
			        if (obj == null)
			            return false;
			        if (getClass() != obj.getClass())
			            return false;
			        «className» other = («className») obj;
			        if (value == null) {
			            if (other.value != null)
			                return false;
			        } else if (!value.equals(other.value))
			            return false;
			        «IF hasIndices»
			        	if (indices == null) {
			        	    if (other.indices != null)
			        	        return false;
			        	} else if (!indices.equals(other.indices))
			        	    return false;
			        «ENDIF»
			        return true;
			    }
			    «misc»
			}
		'''
	}
	
	def static generateCompositeFilter(AttributeDef attribute, String deviceName) 
		'''
			/**
			 * Returns this composite in which a filter was set to the attribute <code>«attribute.name»</code>.
			 * 
			 * @param «attribute.name» The <code>«attribute.name»<code> attribute value to match.
			 * @return this {@link «deviceName»Composite}, filtered using the attribute <code></code>.
			 */
			public «deviceName»Composite and«attribute.name.toFirstUpper»(«attribute.attrTypeRef.fullName» «attribute.name») throws CompositeException {
				filterByAttribute("«attribute.name»", «attribute.name»);
				return this;
			}
		'''
	
	def static generateSubscriptionToSource(SourceDef source, String deviceName)
		'''
			/**
			 * Subscribes to the <code>«source.name»</code> source. After a call to this method, the context will be notified when a
			 * <code>«deviceName»</code> device of this composite publishes a value on its <code>«source.name»</code> source.
			 */
			public void subscribe«source.name.toFirstUpper»() {
				subscribeValue("«source.name»");
			}
			
			/**
			 * Unsubscribes from the <code>«source.name»</code> source. After a call to this method, the context will no more be notified
			 * when a <code>«deviceName»</code> device of this composite publishes a value on its <code>«source.name»</code> source.
			 */
			public void unsubscribe«source.name.toFirstUpper»() {
				unsubscribeValue("«source.name»");
			}
		'''		
	
	def static discovererClassForBehavior(String name, String parentName, String comment, Iterable<DeviceType> devices, Iterable<ContextDef> contexts) '''
        /**
         * An object to discover devices and contexts for the following interaction contract:
         * 
         * <code>
         * «comment.replace("\n", "\n* ")»
         * </code>
         */
        protected final class Discover {
            «IF devices != null»
                «FOR dev : devices»
                    private final «dev.name»DiscovererFor«name» «dev.name.toFirstLower»Discoverer = new «dev.name»DiscovererFor«name»(«parentName».this);
                «ENDFOR»
            «ENDIF»
            «IF contexts != null»
                «FOR ctxt : contexts»
                    private final fr.inria.diagen.core.service.filter.service.ServiceFilter «ctxt.name.toFirstLower»Filter = new fr.inria.diagen.core.service.filter.service.ServiceFilter("«ctxt.classNameSpace»");
                «ENDFOR»
                «FOR ctxt : contexts»
                    
                    private RemoteServiceInfo get«ctxt.name»Rsi() {
                        return getProcessor().discoverComponents(getOutProperties(), «ctxt.name.toFirstLower»Filter).get(0);
                    }
                «ENDFOR»
            «ENDIF»
            «IF devices != null»
                «FOR dev : devices»
                    
                    /**
                     * @return a {@link «dev.name»DiscovererFor«name»} object to discover <code>«dev.name»</code> devices
                     */
                    public «dev.name»DiscovererFor«name» «dev.name.toFirstLower.plural»() {
                        return «dev.name.toFirstLower»Discoverer;
                    }
                «ENDFOR»
            «ENDIF»
            «IF contexts != null»
                «FOR ctxt : contexts»
                    
                    /**
                     * Get the value of the <code>«ctxt.name»</code> context
                     * 
«««                     «FOR ind : ctxt.indices»
«««                     * «paramComment(ind, "the value of the <code>" + ind.name + "</code> index")»
«««                     «ENDFOR»
                     * @return the value of the <code>«ctxt.name»</code> context
                     */
                    public «ctxt.dataTypeRef.fullName» «ctxt.name.toFirstLower»(«ctxt.indices.parameters») {
                        return («ctxt.dataTypeRef.fullName») getProcessor().callGetValue(getOutProperties(), get«ctxt.name»Rsi(), "«ctxt.name»"«
                            if(!ctxt.indices.nullOrEmpty) (",\n        " + ctxt.indices.map[x | x.name.toFirstLower].join(",\n        "))
                        »);
                    }
                «ENDFOR»
            «ENDIF»
        }
    '''
	
     
    def static discovererClassForDevice(DeviceType device, String name, String comment) '''
«««        «device.javadoc(comment)»
        protected final static class «device.name»Discoverer«name» {
            private Service serviceParent;
            
            private «device.name»Discoverer«name»(Service serviceParent) {
                super();
                this.serviceParent = serviceParent;
            }
            
            private «device.name»Composite«name» instantiateComposite() {
                return new «device.name»Composite«name»(serviceParent);
            }
            
            /**
             * Returns a composite of all accessible <code>«device.name»</code> devices
             * 
             * @return a {@link «device.name»Composite«name»} object composed of all discoverable <code>«device.name»</code>
             */
            public «device.name»Composite«name» all() {
                return instantiateComposite();
            }
            
            /**
             * Returns a proxy to one out of all accessible <code>«device.name»</code> devices
             * 
             * @return a {@link «device.name»Proxy«name»} object pointing to a random discoverable <code>«device.name»</code> device
             */
            public «device.name»Proxy«name» anyOne() {
                return all().anyOne();
            }
            «FOR attr : device.allAttributes.sortBy[n | n.name]»
                
                /**
                 * Returns a composite of all accessible <code>«device.name»</code> devices whose attribute <code>«attr.name»</code> matches a given value.
                 * 
«««                 * @param «attr.name» The <code>«attr.name»<code> attribute value to match.«IF !attr.comments.nullOrEmpty» «attr.comments.split("\n").join(" ")»«ENDIF»
                 * @return a {@link «device.name»Composite«name»} object composed of all matching <code>«device.name»</code> devices
                 */
                public «device.name»Composite«name» where«attr.name.toFirstUpper»(«attr.attrTypeRef.fullName» «attr.name») throws CompositeException {
                    return instantiateComposite().and«attr.name.toFirstUpper»(«attr.name»);
                }
            «ENDFOR»
        }
        '''
//        
    def static compositeClassForDevice(DeviceType device, String name, Iterable<SourceDef> registerSources, Iterable<ActionDef> actionExec, String comment) '''
«««        «device.javadoc(comment)»
        protected final static class «device.name»Composite«name» extends fr.inria.diagen.core.service.composite.Composite<«device.name»Proxy«name»> {
            private «device.name»Composite«name»(Service serviceParent) {
                super(serviceParent, "«device.className»");
            }
            
            @Override
            protected «device.name»Proxy«name» instantiateProxy(RemoteServiceInfo rsi) {
                return new «device.name»Proxy«name»(service, rsi);
            }
            «FOR attr : device.allAttributes.sortBy[n | n.name]»
                
                /**
                 * Returns this composite in which a filter was set to the attribute <code>«attr.name»</code>.
                 * 
«««                 * @param «attr.name» The <code>«attr.name»<code> attribute value to match.«IF !attr.comments.nullOrEmpty» «attr.comments.split("\n").join(" ")»«ENDIF»
                 * @return this {@link «device.name»Composite«name»}, filtered using the attribute <code>«attr.name»</code>.
                 */
                public «device.name»Composite«name» and«attr.name.toFirstUpper»(«attr.attrTypeRef.fullName» «attr.name») throws CompositeException {
                    filterByAttribute("«attr.name»", «attr.name»);
                    return this;
                }
            «ENDFOR»
            «IF actionExec != null»
                «FOR action : actionExec»
                    «FOR order : action.orders»
                        
                        /**
«««                         * Executes the <code>«order.stripComments.replace("\n", " ").stripComma»</code> action's method on all devices of this composite.
«««                         «IF !Utils::getComments(order, 1).nullOrEmpty»
                         * 
                         * <p>
«««                         * «Utils::getComments(order, 1).replace("\n", "\n* ")»
«««                         «ENDIF»
«««                         «IF !order.params.nullOrEmpty»
«««                         * 
«««                         «FOR param : order.params»
«««                         * «orderParamComment(order, param)»
«««                         «ENDFOR»
«««                         «ENDIF»
                         */
                        public void «order.name»(«order.params.parameters») throws InvocationException {
                            launchDiscovering();
                            for («device.name»Proxy«name» proxy : proxies) {
                                proxy.«order.name»(«order.params.variableNames»);
                            }
                        }
                    «ENDFOR»
                «ENDFOR»
            «ENDIF»
            «IF registerSources != null»
                «FOR src : registerSources»
                    
                    /**
                     * Subscribes to the <code>«src.name»</code> source. After a call to this method, the context will be notified when a
                     * <code>«device.name»</code> device of this composite publishes a value on its <code>«src.name»</code> source.
                     */
                    public void subscribe«src.name.toFirstUpper»() {
                        subscribeValue("«src.name»");
                    }
                    
                    /**
                     * Unsubscribes from the <code>«src.name»</code> source. After a call to this method, the context will no more be notified
                     * when a <code>«device.name»</code> device of this composite publishes a value on its <code>«src.name»</code> source.
                     */
                    public void unsubscribe«src.name.toFirstUpper»() {
                        unsubscribeValue("«src.name»");
                    }
                «ENDFOR»
            «ENDIF»
        }
     '''
     
     def static proxyClassForDevice(DeviceType device, String name, SourceDef sourceForSubscription, Iterable<SourceDef> sourcesToGet, Iterable<ActionDef> actionsToTrigger, String comment) '''        
        protected final static class «device.name»Proxy«name» extends Proxy {
            private «device.name»Proxy«name»(Service service, RemoteServiceInfo remoteServiceInfo) {
                super(service, remoteServiceInfo);
            }
            «IF sourceForSubscription != null»
            		«sourceForSubscription.generateSubscriptionToSource(device.name)»
           	«ENDIF»
            «IF sourcesToGet != null»
                «FOR source : sourcesToGet»
                    
                    /**
                     * Returns the value of the <code>«source.name»</code> source of this <code>«device.name»</code> device
                     * 
«««                     «IF !source.comments.nullOrEmpty»
                     * <p>
«««                     * «source.comments.replace("\n", "\n* ")»
                     * 
«««                     «ENDIF»
«««                     «FOR ind : source.indices»
«««                     * «paramComment(ind, "the value of the <code>" + ind.name + "</code> index")»
«««                     «ENDFOR»
                     * @return the value of the <code>«source.name»</code> source
                     */
                    public «source.dataTypeRef.fullName» get«source.name.toFirstUpper»(«source.indices.parameters») throws InvocationException {
                        return («source.dataTypeRef.fullName») callGetValue("«source.name»"«if(!source.indices.nullOrEmpty) (", " + source.indices.variableNames)»);
                    }
                «ENDFOR»
            «ENDIF»
            «IF actionsToTrigger != null»
                «FOR action : actionsToTrigger»
                    «FOR order : action.orders»
                        
                        /**
«««                         * Executes the <code>«order.stripComments.replace("\n", " ").stripComma»</code> action's method on this device.
«««                         «IF !Utils::getComments(order, 1).nullOrEmpty»
«««                         * 
«««                         * <p>
«««                         * «Utils::getComments(order, 1).replace("\n", "\n* ")»
«««                         «ENDIF»
«««                         «IF !order.params.nullOrEmpty»
«««                         * 
«««                         «FOR param : order.params»
«««                         * «orderParamComment(order, param)»
«««                         «ENDFOR»
«««                         «ENDIF»
                         */
                        public void «order.name»(«order.params.parameters») throws InvocationException {
                            callOrder("«order.name»"«IF !order.params.nullOrEmpty», «order.params.variableNames»«ENDIF»);
                        }
                    «ENDFOR»
                «ENDFOR»
            «ENDIF»
            «FOR attribute : device.allAttributes.sortBy([attribute | attribute.name])»
                
                /**
«««                 «IF !attr.comments.nullOrEmpty»
«««                 * «attr.comments.split("\n").join("* ")»
«««                 * 
«««                 «ENDIF»
                 * @return the value of the <code>«attribute.name»</code> attribute
                 */
                public «attribute.attrTypeRef.fullName» «attribute.name»() {
                    return («attribute.attrTypeRef.fullName») callGetValue("«attribute.name»");
                }
            «ENDFOR»
        }
	'''
	
//    def static proxyClassForDevice(DeviceDef device, String name, Iterable<SourceDef> discoverableSources, Iterable<SourceDef> sourceGet, Iterable<ActionDef> actionExec, String comment) '''
//        «device.javadoc(comment)»
//        protected final static class «device.name»Proxy«name» extends Proxy {
//            private «device.name»Proxy«name»(Service service, RemoteServiceInfo remoteServiceInfo) {
//                super(service, remoteServiceInfo);
//            }
//            «IF discoverableSources != null»
//                «FOR src : discoverableSources.sort[n1,n2 | n1.name.compareTo(n2.name)]»
//                    
//                    /**
//                     * Subscribes to the <code>«src.name»</code> source. After a call to this method, the context will be notified when the
//                     * <code>«device.name»</code> device of this proxy publishes a value on its <code>«src.name»</code> source.
//                     */
//                    public void subscribe«src.name.toFirstUpper»() {
//                        subscribeValue("«src.name»");
//                    }
//                    
//                    /**
//                     * Unsubscribes from the <code>«src.name»</code> source. After a call to this method, the context will no more be notified
//                     * when the <code>«device.name»</code> device of this proxy publishes a value on its <code>«src.name»</code> source.
//                     */
//                    public void unsubscribe«src.name.toFirstUpper»() {
//                        unsubscribeValue("«src.name»");
//                    }
//                «ENDFOR»
//            «ENDIF»
//            «IF sourceGet != null»
//                «FOR src : sourceGet»
//                    
//                    /**
//                     * Returns the value of the <code>«src.name»</code> source of this <code>«device.name»</code> device
//                     * 
//                     «IF !src.comments.nullOrEmpty»
//                     * <p>
//                     * «src.comments.replace("\n", "\n* ")»
//                     * 
//                     «ENDIF»
//                     «FOR ind : src.indices»
//                     * «paramComment(ind, "the value of the <code>" + ind.name + "</code> index")»
//                     «ENDFOR»
//                     * @return the value of the <code>«src.name»</code> source
//                     */
//                    public «src.dataTypeRef.fullname» get«src.name.toFirstUpper»(«src.indices.parameters») throws InvocationException {
//                        return («src.dataTypeRef.fullname») callGetValue("«src.name»"«if(!src.indices.nullOrEmpty) (", " + src.indices.variableNames)»);
//                    }
//                «ENDFOR»
//            «ENDIF»
//            «IF actionExec != null»
//                «FOR action : actionExec»
//                    «FOR order : action.orders»
//                        
//                        /**
//                         * Executes the <code>«order.stripComments.replace("\n", " ").stripComma»</code> action's method on this device.
//                         «IF !Utils::getComments(order, 1).nullOrEmpty»
//                         * 
//                         * <p>
//                         * «Utils::getComments(order, 1).replace("\n", "\n* ")»
//                         «ENDIF»
//                         «IF !order.params.nullOrEmpty»
//                         * 
//                         «FOR param : order.params»
//                         * «orderParamComment(order, param)»
//                         «ENDFOR»
//                         «ENDIF»
//                         */
//                        public void «order.name»(«order.params.parameters») throws InvocationException {
//                            callOrder("«order.name»"«IF !order.params.nullOrEmpty», «order.params.variableNames»«ENDIF»);
//                        }
//                    «ENDFOR»
//                «ENDFOR»
//            «ENDIF»
//            «FOR attr : device.allAttributes.sortBy([n | n.name])»
//                
//                /**
//                 «IF !attr.comments.nullOrEmpty»
//                 * «attr.comments.split("\n").join("* ")»
//                 * 
//                 «ENDIF»
//                 * @return the value of the <code>«attr.name»</code> attribute
//                 */
//                public «attr.dataTypeRef.fullname» «attr.name»() {
//                    return («attr.dataTypeRef.fullname») callGetValue("«attr.name»");
//                }
//            «ENDFOR»
//        }
//        '''
//

//    
//    def static String paramComment(VariableDef param, String comment) {
//        val name = param.name;
//        val pcomment = Utils::getComments(param, 1).replace("\n", " ");
//        return '''@param «name» «comment».«IF !pcomment.empty» «pcomment»«ENDIF»'''
//    }
//    def static String orderParamComment(OrderDef order, VariableDef param) {
//        val code = order.stripComments.replace("\n", " ").stripComma;
//        return paramComment(param, '''the <code>«param.name»</code> parameter of the <code>«code»</code> method''')
//    }
}