package fr.inria.phoenix.diasuite.diaswarm.generator.template.context

import fr.inria.phoenix.diasuite.diaswarm.diaswarm.BehaviorPublication
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.ContextDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DeviceType
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.EnumDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.InteractionContractDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.PullInteractionContract
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.PushInteractionContract
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.SourceDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.SourceRef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.StructDef
import fr.inria.phoenix.diasuite.diaswarm.generator.template.DiaSwarmEntity
import fr.inria.phoenix.diasuite.diaswarm.generator.template.IConstant
import java.util.ArrayList
import org.eclipse.emf.ecore.EObject

import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.GenericTemplate.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.Utils.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.context.DiaSwarmContextHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.datatype.DiaSwarmDatatypeHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.deployment.DiaSwarmDeploymentHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.device.DiaSwarmDeviceHelper.*

class DiaSwarmContext extends DiaSwarmEntity {

	def generateContext(ContextDef context) {
		context.abstractClassForContext(context.getPackageName)
	}
	
	def generateJob(ContextDef context, PushInteractionContract contract) {
		context.generateJobClass(context.getPackageName, contract)
	}
	
	def generateAdapter(ContextDef context, PushInteractionContract contract) {
		context.generateAdapterClass(context.getPackageName, contract)
	}
	
	def generateResultsParser(ContextDef context, PushInteractionContract contract) {
		context.generateResultsParserClass(context.getPackageName, contract)
	}
	
	def generateOutputFormat(ContextDef context, PushInteractionContract contract) {
		context.generateOutputFormatClass(context.getPackageName, contract)
	}
	
	def generateRecordWriter(ContextDef context, PushInteractionContract contract) {
		context.generateRecordWriterClass(context.getPackageName, contract)
	}
	
//	def generateCustomWritableType(ContextDef context,DataTypeRef dataType, PushInteractionContract contract) {
//		dataType.generateWritableClass(context.getPackageName, contract)
//	}

	def contextIndices(ContextDef context) {
        return context.indices.classForIndices(context.packageName, context.name, context,
            '''Container for the indices of the context <code>«context.name»</code>.'''
        );
    }
	
	def abstractClassForContext(ContextDef context, String packageName) ''' 
        package «packageName»;
        
        «context.imports»
«««        «context.behaviors.filter[interaction.equals(IConstant.PERIODIC_INTERACTION)].findFirst[].mapReduceTypes»
        @SuppressWarnings("all")
        public abstract class Abstract«context.name.toFirstUpper» extends Service «context.mapReduce» {
        
        	/**================== CLASS ATTRIBUTES ==================**/
        	«context.generateAttributes»
        	/*=======================================================*/
        	
        	
        	/**================= CONTEXT CONSTRUCTOR ================**/
        	«context.generateConstructor»
            /*=======================================================*/
            
            
            /**============== METHODS FOR CONTEXT VALUE =============**/
        	«context.generateMethods»
            /*=======================================================*/
            
            
            /**============ CONTEXT RETURN VALUE CLASS ==============**/
            «context.generateReturnValueCode»
            /*=======================================================*/
            
            /**============ INTERACTION CONTRACT METHOD =============**/
            «FOR behavior : context.behaviors»
            	«context.generateInteractionContract(behavior)»
            «ENDFOR»
            /*=======================================================*/
            
            /**============== DEVICE SUBSCRIPTION CODE ==============**/
            «FOR behavior : context.behaviors»
            	«IF !behavior.interaction.equals(IConstant.REQUIRED_INTERACTION)»
            		«val deviceForSubscription = behavior.deviceForSubscription»
«««            		«val providedSource = context.providedSource»
            	
«««            		«FOR providedSource : context.providedSources»
«««            		«IF providedSource != null»
«««						«println("Provided source : " + providedSource.toString)»
«««            			«deviceForSubscription.generateDeviceSubscriptionCode(context.providedSources)»
						«IF deviceForSubscription != null»
							«deviceForSubscription.generateDeviceSubscriptionCode(behavior.providedSource)»
						«ENDIF»
«««            		«ENDFOR»
«««            		«ENDIF»
«««            		«ENDFOR»
            	«ENDIF»
            «ENDFOR»
            /*=======================================================*/
            
            /**================ DEVICE DISCOVERY CODE ===============**/
            
            «FOR behavior : context.interactionContractsForGet»
            	«behavior.discoverCodeForBehavior»
            «ENDFOR»
«««            «IF !context.getDataRequirementsForGet.empty»
«««            	«FOR behavior : context.behaviors»
«««            		«behavior.discoverCodeForBehavior»
«««            	«ENDFOR»
«««            «ENDIF»
            /*=======================================================*/
            
            /**=================== SERVICE METHODS ==================**/
            «context.serviceClassMethods»
            /*=======================================================*/
        }
    '''
    
    def generateJobClass(ContextDef context, String packageName, PushInteractionContract contract) '''
    package «packageName»;
    
    import java.io.IOException;
    import java.util.Iterator;
    
    import org.apache.hadoop.conf.Configured;
    import org.apache.hadoop.fs.Path;
    import org.apache.hadoop.io.LongWritable;
    import org.apache.hadoop.io.Text;
    import org.apache.hadoop.mapred.FileInputFormat;
    import org.apache.hadoop.mapred.FileOutputFormat;
    import org.apache.hadoop.mapred.JobClient;
    import org.apache.hadoop.mapred.JobConf;
    import org.apache.hadoop.mapred.MapReduceBase;
    import org.apache.hadoop.mapred.Mapper;
    import org.apache.hadoop.mapred.OutputCollector;
    import org.apache.hadoop.mapred.Reducer;
    import org.apache.hadoop.mapred.Reporter;
    import org.apache.hadoop.mapred.TextInputFormat;
    import org.apache.hadoop.mapred.TextOutputFormat;
    import org.apache.hadoop.util.Tool;
    import org.apache.log4j.Level;
    import org.apache.log4j.Logger;
    
    import «contract.providedData.groupingAttribute.attrTypeRef.getTypeName»;
    «IF (contract.providedData.mapOutKeyType != null)»
    import «contract.providedData.mapOutKeyType.fullName»;
    «IF (contract.providedData.mapOutKeyType.primitiveTypeRef != null)»
    import «contract.providedData.mapOutKeyType.getHadoopTypeName»;
    «ENDIF»
    «ENDIF»
    «IF (contract.providedData.mapOutValueType != null)»
    import «contract.providedData.mapOutValueType.fullName»;
    «IF (contract.providedData.mapOutValueType.primitiveTypeRef != null)»
    import «contract.providedData.mapOutValueType.getHadoopTypeName»;
	«ENDIF»
	«ENDIF»
    
    «IF (contract.providedData.reduceType != null)»
    import «contract.providedData.reduceType.fullName»;
    «IF (contract.providedData.reduceType.primitiveTypeRef != null)»
    import «contract.providedData.reduceType.getHadoopTypeName»;    
    «ENDIF»
    «ENDIF»
    
    «val mapKeyOutputTypeHadoop = if (contract.providedData.mapOutKeyType != null) contract.providedData.mapOutKeyType.hadoopTypeName.split('\\.').last else "Text"»
    
        
    import fr.inria.diagen.core.mapreduce.MapReduce;
    import fr.inria.diagen.log.DiaLog;
    
    public class «context.name.toFirstUpper»Job extends Configured implements Tool {
    	
    	public static «context.name.toFirstUpper»Adapter jobLauncher;

    	public «context.name.toFirstUpper»Job(
    			«IF contract.providedData.map && contract.providedData.reduce»
    			MapReduce<«(context.mapReduceBehaviors.last.contract as PushInteractionContract).mapReduceTypes»> context) {
    			«ELSEIF contract.providedData.map && !contract.providedData.reduce»
    			fr.inria.diagen.core.mapreduce.Map<«contract.mapTypes»> context) {
    			«ELSEIF !contract.providedData.map && contract.providedData.reduce»
    			fr.inria.diagen.core.mapreduce.Reduce<«contract.reduceTypes»> context) {
				«ENDIF»	
    			
«««    			Class<? extends Abstract«context.name.toFirstUpper»> contextClazz) {
    		try {
    			jobLauncher = new «context.name.toFirstUpper»Adapter(context);
«««    					.getConstructor(ServiceConfiguration.class).newInstance(new Object[] {null}));
    		} catch (Exception e) {
    			System.err.println(e.getMessage());
    		}
    	}
    	
    	«IF contract.providedData.map»
    	public static class «context.name.toFirstUpper»Map extends MapReduceBase implements
    			Mapper<LongWritable, Text, «mapKeyOutputTypeHadoop», «contract.providedData.mapOutValueType.getHadoopTypeName.split('\\.').last»> {
    				
    		@Override
    		public void map(LongWritable key, Text value,
    				OutputCollector<«mapKeyOutputTypeHadoop», «contract.providedData.mapOutValueType.getHadoopTypeName.split('\\.').last»> output, Reporter reporter)
    				throws IOException {
    			jobLauncher.doMap(key, value, output);
    		}
    	}
    	«ENDIF»
    	
    	«IF contract.providedData.reduce»
    	public static class «context.name.toFirstUpper»Reduce extends MapReduceBase
    			implements Reducer<«mapKeyOutputTypeHadoop», «contract.providedData.getMapOutputValueType.getHadoopTypeName.split('\\.').last», «mapKeyOutputTypeHadoop», «contract.providedData.getReduceOutputValueType.getHadoopTypeName.split('\\.').last»> {
«««				implements Reducer<Text, «contract.providedData.getMapOutputValueType.getHadoopTypeName.split('\\.').last», Text, «contract.providedData.getReduceOutputValueType.getHadoopTypeName.split('\\.').last»> {
    		@Override
    		public void reduce(«mapKeyOutputTypeHadoop» key, Iterator<«contract.providedData.mapOutValueType.getHadoopTypeName.split('\\.').last»> values,
    				OutputCollector<«mapKeyOutputTypeHadoop», «contract.providedData.getReduceOutputValueType.getHadoopTypeName.split('\\.').last»> output, Reporter reporter)
    				throws IOException {
    			jobLauncher.doReduce(key, values, output);
    		}
    	}
    	«ENDIF»
    
    	@Override
    	public int run(String[] args) throws Exception {
    		Logger.getRootLogger().getLoggerRepository().resetConfiguration();
    		Logger.getRootLogger().addAppender(
    				DiaLog.getFileAppender("«context.name.toFirstUpper»",
    						"«context.name».log", Level.DEBUG, true));
    	
    		JobConf conf = new JobConf(getConf(), «context.name.toFirstUpper»Job.class);
    		conf.setJobName("«context.name.toFirstUpper»");
    	
    		/* set mapred output types */
    		conf.setOutputKeyClass(Text.class);
    		conf.setOutputValueClass(«contract.providedData.getReduceOutputValueType.getHadoopTypeName.split('\\.').last».class);
    	
    		/* set map output types */
    		conf.setMapOutputKeyClass(«mapKeyOutputTypeHadoop».class);
    		conf.setMapOutputValueClass(«contract.providedData.getMapOutputValueType.getHadoopTypeName.split('\\.').last».class);
    		
    		«IF contract.providedData.map»
			conf.setMapperClass(«context.name.toFirstUpper»Map.class);
    		«ELSE»
    		conf.setNumMapTasks(0);
    		«ENDIF»
    		«IF contract.providedData.reduce»
    		conf.setReducerClass(«context.name.toFirstUpper»Reduce.class);
    		«ELSE»
    		conf.setNumReduceTasks(0);
    		«ENDIF»
    	
    		conf.setInputFormat(TextInputFormat.class);
    		conf.setOutputFormat(«context.name»OutputFormat.class);
    	
    		FileInputFormat.setInputPaths(conf, new Path(args[0]));
    		FileOutputFormat.setOutputPath(conf, new Path(args[1]));
    	
    		JobClient.runJob(conf);
    		return 0;
    	}
    }
    '''
    
    def generateResultsParserClass(ContextDef context, String packageName, PushInteractionContract contract) '''
    «val mapOutKeyType = if (contract.providedData.mapOutKeyType != null) contract.providedData.mapOutKeyType else contract.providedData.groupingAttribute.attrTypeRef»
    «val mapOutKeyTypeName = mapOutKeyType.typeName»
    «val mapOutKeySimpleTypeName = mapOutKeyTypeName.split('\\.').last»
	«val reduceOutValueType = if (contract.providedData.reduce) contract.providedData.reduceType else contract.providedData.mapOutValueType»
    «val reduceOutValueTypeName = reduceOutValueType.typeName»
    «val reduceOutValueSimpleTypeName = reduceOutValueTypeName.split('\\.').last»
	package «packageName»;
	
	import java.io.EOFException;
	import java.io.IOException;
	import java.io.ObjectInputStream;
	import java.util.HashMap;
	«IF (!contract.providedData.reduce)»import java.util.List;
	import java.util.ArrayList;«ENDIF»
	import java.util.Map;
	import fr.inria.diagen.core.mapreduce.ParseMapReduceResults;
	import «mapOutKeyTypeName»;
	import «reduceOutValueTypeName»;
	
	public class «context.name.toFirstUpper»ResultsParser implements ParseMapReduceResults {
		
		@Override
		public Map<?, ?> parseResults(ObjectInputStream resultsInputStream) {
		«IF(!contract.providedData.reduce)»
		Map<«mapOutKeySimpleTypeName», List<«reduceOutValueSimpleTypeName»>> results = new HashMap<«mapOutKeySimpleTypeName», List<«reduceOutValueSimpleTypeName»>>();
		«ELSE»
		Map<«mapOutKeySimpleTypeName», «reduceOutValueSimpleTypeName»> results = new HashMap<«mapOutKeySimpleTypeName», «reduceOutValueSimpleTypeName»>();
		«ENDIF»
	
		try {
			for (;;) {
					Object obj = resultsInputStream.readObject();
					«IF (mapOutKeyType.type instanceof EnumDef)»
					«mapOutKeySimpleTypeName» key = «mapOutKeySimpleTypeName».valueOf(obj.toString());
					«ELSE»
					«mapOutKeySimpleTypeName» key = («mapOutKeySimpleTypeName») obj;
					«ENDIF»
					«IF (reduceOutValueType.type instanceof EnumDef)»
					«reduceOutValueSimpleTypeName» value = «reduceOutValueSimpleTypeName».valueOf(resultsInputStream.readObject());
					«ELSE»
					«reduceOutValueSimpleTypeName» value = («reduceOutValueSimpleTypeName») resultsInputStream.readObject();
					«ENDIF»
					
					«IF(!contract.providedData.reduce)»
					if (results.containsKey(key))
						results.get(key).add(value);
					else {
						List<«reduceOutValueSimpleTypeName»> «reduceOutValueSimpleTypeName.toFirstLower»List = new ArrayList<«reduceOutValueSimpleTypeName»>();
						«reduceOutValueSimpleTypeName.toFirstLower»List.add(value);
						results.put(key, «reduceOutValueSimpleTypeName.toFirstLower»List);
					}
					«ELSE»
					results.put(key, value);
    				«ENDIF»
    			}
    		} catch (EOFException exc) {
    			try {
    				resultsInputStream.close();
    			} catch (IOException e) {
    				e.printStackTrace();
    			}
    
    			return results;
    
    		} catch (ClassNotFoundException e) {
    			e.printStackTrace();
    		} catch (IOException e) {
    			e.printStackTrace();
    		}
    		return null;
    
    	}
    }
    '''
    def generateOutputFormatClass(ContextDef context, String packageName, PushInteractionContract contract) '''
        «val mapOutKeyTypeName = if (contract.providedData.mapOutKeyType != null) contract.providedData.mapOutKeyType.hadoopTypeName else contract.providedData.groupingAttribute.attrTypeRef.hadoopTypeName»
        «val reduceOutValueTypeName = if (contract.providedData.reduce) contract.providedData.reduceType.hadoopTypeName else contract.providedData.mapOutValueType.hadoopTypeName»
        «val mapOutKeySimpleTypeName = mapOutKeyTypeName.split('\\.').last»
        «val reduceOutValueSimpleTypeName = reduceOutValueTypeName.split('\\.').last»
    package «packageName»;
    
    import java.io.IOException;
    
    import org.apache.hadoop.fs.FSDataOutputStream;
    import org.apache.hadoop.fs.FileSystem;
    import org.apache.hadoop.fs.Path;
    import «mapOutKeyTypeName»;
    import «reduceOutValueTypeName»;
    import org.apache.hadoop.mapred.JobConf;
    import org.apache.hadoop.mapred.TextOutputFormat;
    import org.apache.hadoop.util.Progressable;
    
    public class «context.name»OutputFormat extends TextOutputFormat<«mapOutKeySimpleTypeName», «reduceOutValueSimpleTypeName»> {
    	
    	@Override
    	public org.apache.hadoop.mapred.RecordWriter<«mapOutKeySimpleTypeName», «reduceOutValueSimpleTypeName»> getRecordWriter(FileSystem ignored,
    			JobConf job, String name, Progressable progress) throws IOException {
    	
    			// get the current path
    			Path path = new Path("output_«context.name»");
    	
    			// create the full path with the output directory plus our filename
    			Path fullPath = new Path(path, "results_«context.name»");
    	
    			// create the file in the file system
    			FileSystem fs = path.getFileSystem(job);
    			FSDataOutputStream fileOut = fs.create(fullPath, progress);
    	
    			// create our record writer with the new file
    			return new «context.name»RecordWriter(fileOut);
    		}
    }
    '''
    
    def generateRecordWriterClass(ContextDef context, String packageName, PushInteractionContract contract) '''
    «val mapOutKeyType = if (contract.providedData.mapOutKeyType != null) contract.providedData.mapOutKeyType.hadoopTypeName else contract.providedData.groupingAttribute.attrTypeRef.hadoopTypeName»
    «val reduceOutValueType = if (contract.providedData.reduce) contract.providedData.reduceType.hadoopTypeName else contract.providedData.mapOutValueType.hadoopTypeName»
    package «packageName»;
    import java.io.IOException;
    import java.io.ObjectOutputStream;
    
    import org.apache.hadoop.fs.FSDataOutputStream;
    import «mapOutKeyType»;
    import «reduceOutValueType»;

    import org.apache.hadoop.mapred.Reporter;
    
    public class «context.name.toFirstUpper»RecordWriter implements org.apache.hadoop.mapred.RecordWriter<«mapOutKeyType», «reduceOutValueType»> {
    	
    	private ObjectOutputStream out;
    
    	public «context.name.toFirstUpper»RecordWriter(FSDataOutputStream fileOut) {
    		try {
    			out = new ObjectOutputStream(fileOut);
    		} catch (IOException e) {
    			e.printStackTrace();
    		}
    	}
    
    	@Override
    	public void close(Reporter arg0) throws IOException {
    		out.close();
    	}
    
    	@Override
    	public void write(«mapOutKeyType» arg0, «reduceOutValueType» arg1) throws IOException {
    		«IF mapOutKeyType.endsWith("Text")»
    		out.writeObject(arg0.toString());
    		«ELSE»
    		out.writeObject(arg0.get());
    		«ENDIF»
    		«IF reduceOutValueType.endsWith("Text")»
    		out.writeObject(arg1.toString());
    		«ELSE»
    		out.writeObject(arg1.get());
    		«ENDIF»
    	}
    
    }
    '''
    
    def generateAdapterClass(ContextDef context, String packageName, PushInteractionContract contract) '''
    package «packageName»;
    
    import java.io.IOException;
    import java.util.ArrayList;
    import java.util.Iterator;
    import java.util.List;
    
    import org.apache.hadoop.io.LongWritable;
    import org.apache.hadoop.io.Text;
    import org.apache.hadoop.mapred.OutputCollector;
    
    import fr.inria.diagen.core.mapreduce.MapCollector;
    import fr.inria.diagen.core.mapreduce.MapReduce;
    import fr.inria.diagen.core.mapreduce.ReduceCollector;
    import fr.inria.diagen.core.mapreduce.Utils;
    
    import «contract.providedData.groupingAttribute.attrTypeRef.getTypeName»;
    «IF (contract.providedData.mapOutKeyType != null)»
    import «contract.providedData.mapOutKeyType.fullName»;
    «IF (contract.providedData.mapOutKeyType.primitiveTypeRef != null)»
    import «contract.providedData.mapOutKeyType.getHadoopTypeName»;
    «ENDIF»
    «ENDIF»
    «IF (contract.providedData.mapOutValueType != null)»
    import «contract.providedData.mapOutValueType.fullName»;
    «IF (contract.providedData.mapOutValueType.primitiveTypeRef != null)»
    import «contract.providedData.mapOutValueType.getHadoopTypeName»;
	«ENDIF»
    «ENDIF»
    «IF (contract.providedData.reduceType != null)»
    import «contract.providedData.reduceType.fullName»;
    «IF (contract.providedData.reduceType.primitiveTypeRef != null)»
    import «contract.providedData.reduceType.getHadoopTypeName»;
    «ENDIF»
    «ENDIF»
    
    public class «context.name.toFirstUpper»Adapter {
    	
    	public «context.name.toFirstUpper»Adapter(
    			«IF contract.providedData.map && contract.providedData.reduce»
    			MapReduce<«contract.mapReduceTypes»> adaptee) {
    			«ELSEIF contract.providedData.map && !contract.providedData.reduce»
    			fr.inria.diagen.core.mapreduce.Map<«contract.mapTypes»> adaptee) {
    			«ELSEIF !contract.providedData.map && contract.providedData.reduce»
    			fr.inria.diagen.core.mapreduce.Reduce<«contract.reduceTypes»> adaptee) {
    			«ENDIF»	
    		this.adaptee = adaptee;
    	}
    	
    	private «IF contract.providedData.map && contract.providedData.reduce»
    	    	MapReduce<«contract.mapReduceTypes»> adaptee;
    	    	«ELSEIF contract.providedData.map && !contract.providedData.reduce»
    	    	fr.inria.diagen.core.mapreduce.Map<«contract.mapTypes»> adaptee;
    	    	«ELSEIF !contract.providedData.map && contract.providedData.reduce»
    	    	fr.inria.diagen.core.mapreduce.Reduce<«contract.reduceTypes»> adaptee;
    	    	«ENDIF»	
		
    	«val groupingAttribute = contract.providedData.groupingAttribute»
		«val sourceName = contract.providedData.dataSource.sourceRef.source.name»
		«val mapKeyInputType = groupingAttribute.attrTypeRef»
    	«val mapKeyInputTypeName = groupingAttribute.attrTypeRef.fullName.split('\\.').last»
    	«val mapValueInputType = contract.providedData.dataSource.sourceRef.source.dataTypeRef»
    	«val mapValueInputTypeName = contract.providedData.dataSource.sourceRef.source.dataTypeRef.fullName.split('\\.').last»
		«««		«val mapValueInputType = contract.providedData.dataSource.sourceRef.source.dataTypeRef.fullName.split('\\.').last»
		«val mapKeyOutputTypeDefined = if(contract.providedData.mapOutKeyType != null) true else false»
    	«var mapKeyOutputType = if(mapKeyOutputTypeDefined) contract.providedData.mapOutKeyType.fullName.split('\\.').last else mapKeyInputTypeName»
    	«var mapKeyOutputTypeHadoop = if(mapKeyOutputTypeDefined) contract.providedData.mapOutKeyType.getHadoopTypeName.split('\\.').last else groupingAttribute.attrTypeRef.hadoopTypeName.split('\\.').last»
		«val mapValueOutputType = contract.providedData.mapOutValueType.typeName.split('\\.').last»
    	«val mapValueOutputTypeHadoop = contract.providedData.mapOutputValueType.getHadoopTypeName.split('\\.').last»
		«««    	    		«val mapValueOutputType = contract.providedData.mapType.literal.split('\\.').last»
    	«val String reduceKeyOutputType = mapKeyOutputType»
    	«val String reduceKeyOutputTypeHadoop = mapKeyOutputTypeHadoop»
    	«var String reduceValueOutputType = contract.providedData.reduceOutputValueType.typeName.split('\\.').last»
    	«var String reduceValueOutputTypeHadoop =  contract.providedData.reduceOutputValueType.hadoopTypeName.split('\\.').last»
    	«««    		«var String reduceValueOutPutType = contract.providedData.reduceType.literal.split('\\.').last»
    	
    	«IF contract.providedData.map»
    	public void doMap(LongWritable key, Text value,
    			final OutputCollector<«mapKeyOutputTypeHadoop», «mapValueOutputTypeHadoop»> output) {
    		
    		/*********** transform hadoop types to diaswarm ***********/
    		«IF (mapKeyInputType.type) != null && (mapKeyInputType.type instanceof StructDef)»
    		«FOR field : (mapKeyInputType.type as StructDef).structFields»
    		«IF field instanceof StructDef»
    		«field.dataTypeRef.typeName.split('\\.').last» «field.name» = Utils.getFieldFromJSON("«field.name»", value);
    		«ELSE»
    		«field.dataTypeRef.typeName» «field.name» = «field.dataTypeRef.typeName.split('\\.').last».valueOf(Utils.getFieldFromJSON("«field.name»", value));
    		«ENDIF»
    		«ENDFOR»
    		«mapKeyInputTypeName.split('\\.').last» «groupingAttribute.name»Key = new «mapKeyInputTypeName.split('\\.').last»(«(mapKeyInputType.type as StructDef).structFields.map[field | field.name].join(", ")»);
    		«ELSE»
    		«mapKeyInputTypeName.split('\\.').last» «groupingAttribute.name»Key = «mapKeyInputTypeName.split('\\.').last».valueOf(Utils
    		    				.getFieldFromJSON("«groupingAttribute.name»", value));
    		«ENDIF»
    		
    		«IF (mapValueInputType.type) != null && (mapValueInputType.type instanceof StructDef)»
    		«FOR field : (mapValueInputType.type as StructDef).structFields»
    		«IF field instanceof StructDef»
    		«field.dataTypeRef.typeName» «field.name» = Utils.getFieldFromJSON("«field.name»", value);
    		«ELSE»
    		«field.dataTypeRef.typeName» «field.name» = «field.dataTypeRef.typeName».valueOf(Utils.getFieldFromJSON("«field.name»", value));
    		«ENDIF»
    		«ENDFOR»
    		«mapValueInputTypeName.split('\\.').last» «sourceName»Value = new «mapValueInputTypeName.split('\\.').last»(«(mapValueInputType.type as StructDef).structFields.map[field | field.name].join(", ")»);
    		«ELSE»
    		«mapValueInputTypeName.split('\\.').last» «sourceName»Value = «mapValueInputTypeName.split('\\.').last».valueOf(Utils
    				.getFieldFromJSON("«sourceName»", value));
    		«ENDIF»
    		String[] args = new String[] { Utils
					.getFieldFromJSON("position", value) };
    		/**********************************************************/
    		
    		adaptee.map(«groupingAttribute.name»Key, «sourceName»Value,
    				new MapCollector<«mapKeyOutputType», «mapValueOutputType»>() {
    					
    					@Override
    					public void emitMap(«mapKeyOutputType» key, «mapValueOutputType» value) {
«««    						Text mapKeyOut = Utils.getTextWritable(key.toString());
							«mapKeyOutputTypeHadoop» mapKeyOut = «if(mapKeyOutputTypeHadoop.equals("Text")) "new Text(key.toString());" else "new " + mapKeyOutputTypeHadoop + "(key);"»
							«mapValueOutputTypeHadoop» mapValueOut = «if(mapValueOutputTypeHadoop.equals("Text")) "new Text(value.toString());" else "new " + mapValueOutputTypeHadoop + "(value);"»
«««    						«contract.providedData.mapOutValueType.getHadoopTypeName.split('\\.').last» mapValueOut = Utils
«««    								.get«contract.providedData.mapOutValueType.getHadoopTypeName.split('\\.').last»(value);
    								
    						try {
    							output.collect(mapKeyOut, mapValueOut);
    						} catch (IOException e) {
    							e.printStackTrace();
    						}
    					}
    				}, args);
    	}
    	«ENDIF»
    	
    	«IF contract.providedData.reduce»
    	public void doReduce(«mapKeyOutputTypeHadoop» key, Iterator<«mapValueOutputTypeHadoop»> values,
    			final OutputCollector<«reduceKeyOutputTypeHadoop», «reduceValueOutputTypeHadoop»> output) {
    		
    		/*********** transform hadoop types to diaswarm ***********/
    		«mapKeyOutputType» «groupingAttribute.name» = «if(mapKeyOutputTypeHadoop.equals("Text")) mapKeyOutputType + ".valueOf(key.toString());" else "key.get();"»
«««    		List<«mapValueOutputType»> «sourceName»Values = Utils.toList(values);
«««    		
«««    		while (values.hasNext()) {
«««    			«mapValueOutputType» «sourceName» = values.next().get();
«««    			«sourceName»Values.add(«sourceName»);
«««    		}
    		List<«mapValueOutputType»> «sourceName»Values = new ArrayList<«mapValueOutputType»>();
    		 
    		while (values.hasNext()) {
    			«if(mapValueOutputTypeHadoop.equals("Text")) sourceName + "Values.add(" + mapValueOutputType + ".valueOf(" + "values.next().toString()));" else sourceName + "Values.add(values.next().get());" »
    		}
    		/**********************************************************/
    		adaptee.reduce(«groupingAttribute.name», «sourceName»Values,
    				new ReduceCollector<«reduceKeyOutputType», «reduceValueOutputType»>() {
    					
    					@Override
    					public void emitReduce(«reduceKeyOutputType» key, «reduceValueOutputType» value) {
    						«reduceKeyOutputTypeHadoop» reduceKeyOut = «if(reduceKeyOutputTypeHadoop.equals("Text")) "new Text(key.toString());" else "new " + reduceKeyOutputTypeHadoop + "(key);"»
    						«reduceValueOutputTypeHadoop» reduceValueOut = «if(reduceValueOutputTypeHadoop.equals("Text")) "new Text(value.toString());" else "new " + reduceValueOutputTypeHadoop + "(value);"»
«««    						Text reduceKeyOut = Utils.getTextWritable(key
«««    								.toString());
«««    						«contract.providedData.getReduceOutputValueType.hadoopTypeName.split('\\.').last» reduceValueOut = Utils
«««    								.get«contract.providedData.getReduceOutputValueType.hadoopTypeName.split('\\.').last»(value);
    								
    						try {
    							output.collect(reduceKeyOut, reduceValueOut);
    						} catch (IOException e) {
    							// TODO Auto-generated catch block
    							e.printStackTrace();
    						}
    					}
    				});
    	}
    	«ENDIF»
    }
    '''

//    /** ================== CLASS ATTRIBUTES ================== **/
//	private java.lang.String contextValue;
//	private static final Map<String, Object> properties;
//	private static final Map<String, Object> deliveryModels;
//
//	static {
//		deliveryModels = new HashMap<String, Object>();
//		deliveryModels.put(
//				"testSource",
//				new DataDeliveryModel.Builder()
//						.deliveryModel(DataDeliveryEnum.PERIODIC).period(10)
//						.timeUnit(TimeUnit.S).build());
//
//		properties = new HashMap<String, Object>();
//		properties.put(ServiceProperties.DELIVERY.toString(), deliveryModels);
//	}
	
    def generateAttributes(ContextDef context)  '''
    	«IF context.indices.nullOrEmpty»
    		private «context.dataTypeRef.fullName» contextValue;
    	«ELSE»
    		private java.util.Map<«context.name»Indices, «context.dataTypeRef.fullName»> contextValues;
    	«ENDIF»
    	
    		private static final Map<String, Object> deliveryModels;
    		
    		static {
    			deliveryModels = new HashMap<String, Object>();
    			«FOR behavior : context.behaviors»
    			«IF behavior.interaction.equals(IConstant.PERIODIC_INTERACTION)»
    				«val periodicity = (behavior.contract as PushInteractionContract).providedData.dataSource.periodicity»
    				«val period = (behavior.contract as PushInteractionContract).providedData.dataSource.periodicity.value»
«««    				«val period = (behavior.contract as PushInteractionContract).providedData.dataSource.sourceRef.periodicity.value»
    				«val timeUnit = (behavior.contract as PushInteractionContract).providedData.dataSource.periodicity.timeUnit.getName()»
    				«val groupingPeriod = (behavior.contract as PushInteractionContract).providedData.periodicity»
    				«val valuesNumber = if(groupingPeriod != null) (groupingPeriod.periodicityInSeconds/periodicity.periodicityInSeconds) else 1»
    				
    				«IF groupingPeriod != null»
    					«getDeliveryModelsMap
    						(IConstant.PERIODIC_INTERACTION, 
    							(behavior.contract as PushInteractionContract).contextDataSourceName,
    								groupingPeriod.value,
    									groupingPeriod.timeUnit.getName(), valuesNumber)»
    				«ELSE»
    					«getDeliveryModelsMap
    						(IConstant.PERIODIC_INTERACTION, 
    							(behavior.contract as PushInteractionContract).contextDataSourceName,
    								period,
    									timeUnit, valuesNumber)»
    				«ENDIF»
    			«ELSEIF behavior.interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION) && (behavior.contract as PushInteractionContract).providedData.periodicity != null»
    					«getDeliveryModelsMap
    						(IConstant.PERIODIC_INTERACTION, 
    							(behavior.contract as PushInteractionContract).contextDataSourceName,
    								(behavior.contract as PushInteractionContract).providedData.periodicity.value,
    									(behavior.contract as PushInteractionContract).providedData.periodicity.timeUnit.getName(), -1)»
    			«ELSEIF behavior.interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION)»
    					«getDeliveryModelsMap
    						(IConstant.EVENT_DRIVEN_INTERACTION, 
    							(behavior.contract as PushInteractionContract).contextDataSourceName,
    								0, null, 0)»
    			«ENDIF»
    			«ENDFOR»
    		}
    	'''

    def getImports(ContextDef context) '''
        import fr.inria.diagen.core.ServiceConfiguration;
        import fr.inria.diagen.core.exception.InvocationException;
        «IF context.allSources.exists[!device.allAttributes.nullOrEmpty]»
        import fr.inria.diagen.core.service.composite.CompositeException;
        «ENDIF»
        import fr.inria.diagen.core.network.RemoteServiceInfo;
        import fr.inria.diagen.core.service.local.Service;
        import fr.inria.diagen.core.service.proxy.Proxy;
        import fr.inria.diagen.core.service.buffer.Buffer;
        «IF !context.behaviors.filter[(interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION))
        	||(interaction.equals(IConstant.PERIODIC_INTERACTION))].isEmpty»
        	import java.util.HashMap;
        	import java.util.Map;
        	import fr.inria.diagen.core.network.DataDeliveryEnum;
        	import fr.inria.diagen.core.network.DataDeliveryModel;
        	import fr.inria.diagen.core.service.local.ServiceProperties;
        «ENDIF»
        
        «FOR behavior : context.behaviors.filter[(interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION))
        	||(interaction.equals(IConstant.PERIODIC_INTERACTION))]»
		«IF behavior.interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION) || behavior.interaction.equals(IConstant.PERIODIC_INTERACTION)»
        	«val providedSources = context.providedDeviceSources»
        	«val providedContexts = context.providedContexts»
        	
        	«IF providedSources != null»
        		«FOR providedSource : providedSources»        		
        			«val deviceName = providedSource.device.name»
        			«val sourceName = providedSource.source.name»
        			«val sourceClassName = providedSource.device.packageNameForSource + "." + sourceName.toFirstUpper + "From" + deviceName.toFirstUpper»
        	import «sourceClassName»;
        		«ENDFOR»
        	«ENDIF»
        	«IF providedContexts != null»
        		«FOR providedContext : providedContexts»
        			«val contextClassName = providedContext.packageName + "." + providedContext.name.toFirstUpper + "Value"»
        	import «contextClassName»;
        		«ENDFOR»
        	«ENDIF»
        «ENDIF»
		«IF (behavior.interaction.equals(IConstant.PERIODIC_INTERACTION) || (behavior.interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION) && (behavior.contract as PushInteractionContract).providedData.periodicity != null))»
		import java.util.List;
		import fr.inria.diagen.core.network.TimeUnit;
		import fr.inria.diagen.core.service.buffer.DeviceSourceBuffer;
		«ENDIF»
		«val dataTypeRef = (behavior.contract as PushInteractionContract).getDataSource»
		«IF dataTypeRef.isCustomType»
		import «dataTypeRef.typeName»;
		«ENDIF»
		«IF (behavior.contract as PushInteractionContract).hasGroupingAttribute»
		import «(behavior.contract as PushInteractionContract).providedData.groupingAttribute.attrTypeRef.fullName»;
		«ENDIF»
		«IF ((behavior.contract as PushInteractionContract).providedData.map && (behavior.contract as PushInteractionContract).providedData.reduce)»
		import fr.inria.diagen.core.mapreduce.MapReduce;
		«ENDIF»
		«IF ((behavior.contract as PushInteractionContract).providedData.mapOutKeyType != null)»
		import «(behavior.contract as PushInteractionContract).providedData.mapOutKeyType.fullName»;
		«ENDIF»
		«IF ((behavior.contract as PushInteractionContract).providedData.mapOutValueType != null)»
		import «(behavior.contract as PushInteractionContract).providedData.mapOutValueType.fullName»;
		«ENDIF»
		«IF ((behavior.contract as PushInteractionContract).providedData.reduceType != null)»
		import «(behavior.contract as PushInteractionContract).providedData.reduceType.fullName»;
		«ENDIF»
«««		«IF ((behavior.contract as PushInteractionContract).providedData.mapOutValueType != null)»
«««		import «(behavior.contract as PushInteractionContract).providedData.mapOutValueType.fullName»;
«««		«ENDIF»
		
		
		
		«ENDFOR»
    '''
    
    def generateConstructor(ContextDef context) '''		
    	public Abstract«context.name.toFirstUpper»(ServiceConfiguration serviceConfiguration) {
    		super("/Context/«context.name.toFirstUpper»/", serviceConfiguration);
    		«IF !context.indices.nullOrEmpty»
    		contextValues = new java.util.HashMap<«context.name»Indices, «context.dataTypeRef.fullName»>();
    		«ENDIF»
    		
    		
    	«IF !context.behaviors.filter[(interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION))
        	||(interaction.equals(IConstant.PERIODIC_INTERACTION))].empty»
    		addToOutProperties(ServiceProperties.DELIVERY.toString(),
    				deliveryModels);
        «ENDIF»
    	}
    '''
    
    def generateMethods(ContextDef context) '''
    // Code relative to the return value of the context        
        
        private void set«context.name»(«context.dataTypeRef.fullName» newContextValue«if(!context.indices.nullOrEmpty) ", " + context.indices.parameters») {
            «IF !context.indices.nullOrEmpty»
                «context.name»Indices indices = new «context.name»Indices(«context.indices.variableNames»);
                contextValues.put(indices, newContextValue);
            «ELSE»
                contextValue = newContextValue;
            «ENDIF»
            getProcessor().publishValue(getOutProperties(), "«context.name»", newContextValue«if(!context.indices.nullOrEmpty) ", " + context.indices.variableNames»);
        }
        
        /**
         * Get the last value of the context
         * 
         «FOR ind : context.indices»
         * @param «ind.name» the index <code>«ind.name»</code> for the last value
         «ENDFOR»
         * @return the latest value published by the context«if(!context.indices.nullOrEmpty) " for the given indexes"»
         */
        protected final «context.dataTypeRef.fullName» getLastValue(«context.indices.parameters») {
            «IF !context.indices.nullOrEmpty»
                «context.name»Indices indices = new «context.name»Indices(«context.indices.variableNames»);
                return contextValues.get(indices);
            «ELSE»
                return contextValue;
            «ENDIF»
        }
    '''
        
	def generateReturnValueCode(ContextDef context) '''        
		«IF !context.behaviors.filter[behaviorPublication.equals(BehaviorPublication::MAYBE)].empty»
            /**
             * A class that represents a value that might be published for the <code>«context.name»</code> context. It is used by
             * event methods that might or might not publish values for this context.
             */
            protected final static class «context.name»Value {
                
                // The value of the context
                private «context.dataTypeRef.fullName» value;
                «FOR index: context.indices.sortBy[index | index.name]»
                    // The «index.name» index
                    private «index.dataTypeRef.fullName» «index.name»;
                «ENDFOR»
                // Indicates whether the value should be published or not
                private boolean doPublish;
                
                public «context.name»Value(«context.dataTypeRef.fullName» value,«
                        if(!context.indices.nullOrEmpty) ("\n        "
                            + context.indices.parameters
                            + ",\n        "
                        ) else " "
                    »boolean doPublish) {
                    this.value = value;
                    «FOR index: context.indices.sortBy[index | index.name]»
                        this.«index.name» = «index.name»;
                    «ENDFOR»
                    this.doPublish = doPublish;
                }
                
                /**
                 * @return the value of the context that might be published
                 */
                public «context.dataTypeRef.fullName» getValue() {
                    return value;
                }
                
                /**
                 * Sets the value that might be published
                 * 
                 * @param value the value that will be published if {@link #doPublish()} returns true
                 */
                public void setValue(«context.dataTypeRef.fullName» value) {
                    this.value = value;
                }
                «FOR index: context.indices.sortBy[index | index.name]»
                        
                        /**
                         * @return the <code>«index.name»</code> index for the value of the context that might be published
                         */
                        public «index.dataTypeRef.fullName» get«index.name.toFirstUpper»() {
                            return «index.name»;
                        }
                        
                        /**
                         * Sets the <code>«index.name»</code> index for the value that might be published
                         * 
                         * @param «index.name» the <code>«index.name»</code> index
                         */
                        public void set«index.name.toFirstUpper»(«index.dataTypeRef.fullName» «index.name») {
                            this.«index.name» = «index.name»;
                        }
                «ENDFOR»
                
                /**
                 * @return true if the value should be published
                 */
                public boolean doPublish() {
                    return doPublish;
                }
                
                /**
                 * Set the value to be publishable or not
                 * 
                 * @param doPublish if true, the value will be published
                 */
                public void setDoPublish(boolean doPublish) {
                    this.doPublish = doPublish;
                }
            }
        «ENDIF»
        // End of code relative to the return value of the context
    '''   
    
    def generateInteractionContract(ContextDef context, InteractionContractDef interactionContract) {
	        println("HContext : " + context + " has " + context.behaviors.size + " behaviors !")
//        val interactionContract = context.behavior
        val interaction = interactionContract.interaction
        val contract = interactionContract.contract
        val discoveryRequired = serviceDiscoveryRequired(interaction, contract)
        val publicationType = interactionContract.behaviorPublication                  
               
		// Return value
		switch interaction {			
			// when required ...
			case IConstant.REQUIRED_INTERACTION: {
				println("Generating method for when required IC !")
				val indices = context.indices.parameters;				
				val args = indices.getArgumentsForRequiredInteraction(discoveryRequired)
				return '''protected abstract «context.getContractReturnType(interactionContract.behaviorPublication, true)» getValue(«args»);'''					
			}
			// when provided ...			
			case IConstant.EVENT_DRIVEN_INTERACTION: {
				println("Generating method for when provided IC !")		
				return context.generatePushInteractionContract(contract, interaction, publicationType, discoveryRequired)
			}			
			// when periodic ...
			case IConstant.PERIODIC_INTERACTION: {
				println("Generating method for when periodic IC !")
				return context.generatePushInteractionContract(contract, interaction, publicationType, discoveryRequired)				
			}
		}
    }
    
    def generatePushInteractionContract(ContextDef context, EObject contract, String interaction, BehaviorPublication publicationType, Boolean discoveryRequired) {   	
    	val pushContract = (contract as PushInteractionContract)
		val providedSource = pushContract.providedData.dataSource.sourceRef
				
		// context received a device source value
		if(providedSource != null) {
			val sourceName = providedSource.source.name
			val sourceType = providedSource.source.dataTypeRef
			var String sourceFromDevice = providedSource.sourceFromDevice
			
			if(interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION))
				context.generatePushHandlerMethod(contract, interaction, sourceFromDevice, sourceName, sourceFromDevice, publicationType, discoveryRequired)
			else {
				val sourceTypeName = if(sourceType.type != null) sourceType.type.name else sourceType.primitiveTypeRef.literal
				context.generatePushHandlerMethod(contract, interaction, sourceFromDevice, sourceName, sourceTypeName, publicationType, discoveryRequired)
			}

		// context received a context value 
		} else {
			val providedContext = pushContract.providedData.dataSource.context
			val contextName = providedContext.name
			context.generatePushHandlerMethod(contract, interaction, contextName, contextName, contextName + "Value", publicationType, discoveryRequired)
		}				
    }
    
//    def generateSimplePushInteractionContract(ContextDef context, EObject contract, String interaction, BehaviorPublication publicationType, Boolean discoveryRequired) {   	
//    	val pushContract = (contract as PushInteractionContract)
//		val providedSource = pushContract.providedData.dataSource.sourceRef							
//				
//		// context received a device source value
//		if(providedSource != null) {
//			val sourceName = providedSource.source.name
//			var String sourceFromDevice
//			context.generatePushHandlerMethod(sourceFromDevice, sourceName, sourceFromDevice, publicationType, discoveryRequired)
//
//		// context received a context value 
//		} else {
//			val providedContext = pushContract.providedData.dataSource.context 
//			val contextName = providedContext.name
//			context.generatePushHandlerMethod(contextName, contextName, contextName + "Value", publicationType, discoveryRequired)
//		}				
//    }
    
	def generatePushHandlerMethod(ContextDef context, EObject contract, String interaction, String dataSourceName, String providedValueName, String providedValueTypeName, BehaviorPublication publicationType, Boolean discoveryRequired) {
    	val discover = if(discoveryRequired) ", Discover" + " discover" else "";
    	var pushContract = (contract as PushInteractionContract)
		var String groupingAttrName = "id";
		var String groupingAttrType = "String";
		val Boolean groupingPeriodicity = pushContract.providedData.periodicity != null
		var String sourceMap;
		
		if(pushContract.hasGroupingAttribute) {
				val groupingAttribute = pushContract.providedData.groupingAttribute
				groupingAttrName = groupingAttribute.name
				groupingAttrType = if(groupingAttribute.attrTypeRef.type != null)  groupingAttribute.attrTypeRef.type.name else groupingAttribute.attrTypeRef.primitiveTypeRef.literal				
		}
		
		if(pushContract.providedData.map && pushContract.providedData.reduce) {
			sourceMap = '''Map<«groupingAttrType», «pushContract.providedData.reduceType.typeName.split('\\.').last»> «providedValueName.toFirstLower»By«groupingAttrName.toFirstUpper»Map''';
//			sourceMap = '''Map<«groupingAttrType», «pushContract.providedData.reduceType.literal.split('\\.').last»> «providedValueName.toFirstLower»By«groupingAttrName.toFirstUpper»Map''';
		}
		else if(pushContract.providedData.map) {
			sourceMap = '''Map<«groupingAttrType», List<«pushContract.providedData.mapOutValueType.typeName.split('\\.').last»>> «providedValueName.toFirstLower»By«groupingAttrName.toFirstUpper»Map''';
		}
		else if(!groupingPeriodicity) {
			sourceMap = '''Map<«groupingAttrType», List<«providedValueTypeName»>> «providedValueName.toFirstLower»By«groupingAttrName.toFirstUpper»Map''';
		} else {
			sourceMap = '''Map<«groupingAttrType», List<Buffer<«providedValueTypeName»>>> «providedValueName.toFirstLower»By«groupingAttrName.toFirstUpper»Map''';
		}
			
		if(interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION) && (pushContract.providedData.map || pushContract.providedData.reduce))	
			return '''protected abstract «context.getContractReturnType(publicationType, false)» on«dataSourceName»(«sourceMap»«discover»);'''
			
    	if(interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION)) 
			return '''protected abstract «context.getContractReturnType(publicationType, false)» on«dataSourceName»(«providedValueTypeName» «providedValueName.toFirstLower»«discover»);'''
		
		
		
//		else if(pushContract.hasGroupingAttribute) {
//				val groupingAttribute = pushContract.providedData.groupingAttribute
//				groupingAttrName = groupingAttribute.name
//				groupingAttrType = if(groupingAttribute.attrTypeRef.type != null)  groupingAttribute.attrTypeRef.type.name else groupingAttribute.attrTypeRef.primitiveTypeRef.literal				
//		}
		
		
		
		return '''protected abstract «context.getContractReturnType(publicationType, false)» onPeriodic«dataSourceName»(«sourceMap»«discover»);'''
	}
    
	/* Create the code of a discover object to register to sources of a device */
    def static generateDeviceSubscriptionCode(DeviceType device, SourceDef sourceForSubscription) { '''
        protected final «device.name.toFirstUpper»Subscriber «device.name.toFirstLower»Subscriber = new «device.name»Subscriber(this);
        
        «device.generateDeviceSubscriberClass»
        «device.generateDeviceCompositeClass(sourceForSubscription)»
        «device.proxyClassForDevice("", sourceForSubscription, null, null, '''A proxy to one <code>«device.name»</code> device''')»
        // End of discover part for «device.name» devices
    '''
    }    
	
	def static generateDeviceSubscriberClass(DeviceType device) '''        
        protected final static class «device.name»Subscriber {
            private Service serviceParent;
            
            private «device.name»Subscriber(Service serviceParent) {
                super();
                this.serviceParent = serviceParent;
            }
            
            private «device.name»Composite instantiateComposite() {
                return new «device.name»Composite(serviceParent);
            }
            
            /**
             * Returns a composite of all accessible <code>«device.name»</code> devices
             * 
             * @return a {@link «device.name»Composite} object composed of all discoverable <code>«device.name»</code>
             */
            public «device.name»Composite all() {
                return instantiateComposite();
            }
            
            /**
             * Returns a proxy to one out of all accessible <code>«device.name»</code> devices
             * 
             * @return a {@link «device.name»Proxy} object pointing to a random discoverable <code>«device.name»</code> device
             */
            public «device.name»Proxy anyOne() {
                return all().anyOne();
            }
            «FOR attr : device.allAttributes.sortBy[n | n.name]»
                
                /**
                 * Returns a composite of all accessible <code>«device.name»</code> devices whose attribute <code>«attr.name»</code> matches a given value.
                 * 
                 * @return a {@link «device.name»Composite} object composed of all matching <code>«device.name»</code> devices
                 */
                public «device.name»Composite where«attr.name.toFirstUpper»(«attr.attrTypeRef.fullName» «attr.name») throws CompositeException {
                    return instantiateComposite().and«attr.name.toFirstUpper»(«attr.name»);
                }
            «ENDFOR»
        }
	'''
        
	def static generateDeviceCompositeClass(DeviceType device, SourceDef sourceForSubscription) '''       
        protected final static class «device.name»Composite extends fr.inria.diagen.core.service.composite.Composite<«device.name»Proxy> {
            private «device.name»Composite(Service serviceParent) {
                super(serviceParent, "«device.className»");
            }
            
            @Override
            protected «device.name»Proxy instantiateProxy(RemoteServiceInfo rsi) {
                return new «device.name»Proxy(service, rsi);
            }
            «FOR attribute : device.allAttributes.sortBy[attribute | attribute.name]»
            	«attribute.generateCompositeFilter(device.name)»
            «ENDFOR»

            «sourceForSubscription.generateSubscriptionToSource(device.name)»
        }
	'''

    def serviceClassMethods(ContextDef context) '''
        // Methods from the Service class
        @Override
        protected final void internalPostInitialize() {
«««        «IF !context.behavior.interaction.equals(IConstant.REQUIRED_INTERACTION)»
        «FOR providedCtx : context.providedContexts.sortWith[n1, n2 | n1.name.compareTo(n2.name)]»
        	subscribeValue("«providedCtx.name»", "«providedCtx.classNameSpace»"); // subscribe to «providedCtx.name» context
        «ENDFOR»
«««        «ENDIF»
        	postInitialize();
        }
        	
«««        «IF !context.behavior.interaction.equals(IConstant.REQUIRED_INTERACTION)»
        «IF !context.providedSources.empty»
        	@Override protected void postInitialize() {
        		// Default implementation of post initialize: subscribe to all required devices 
        		«FOR providedSource : context.providedSources.sortWith[n1, n2 | n1.name.compareTo(n2.name)]»
        			«providedSource.deviceName.toFirstLower»Subscriber.all().subscribe«providedSource.name.toFirstUpper»();
        			// subscribe to «providedSource.name» from all «providedSource.deviceName» devices
        		«ENDFOR»
        	}
		«ENDIF»
        
        
            @Override
            public final void valueReceived(java.util.Map<String, Object> properties, RemoteServiceInfo source, String eventName, Object value, Object... indexes) {
            	«FOR interactionContract : context.behaviors.filter[!interaction.equals(IConstant.REQUIRED_INTERACTION)]»
            		«val providedContext = (interactionContract.contract as PushInteractionContract).providedData.dataSource.context»
            		«val providedSource = (interactionContract.contract as PushInteractionContract).providedData.dataSource.sourceRef»
            		«val dataRequirements = (interactionContract.contract as PushInteractionContract).pullInteractionContract != null»
            		«val publication = interactionContract.behaviorPublication»
            		«val interaction = interactionContract.interaction»
            		
            		«IF providedContext != null»
            			«valueReceivedFromContext(interactionContract, context, interaction, providedContext, publication, dataRequirements)»
            		«ELSE»
            			«valueReceivedFromDevice(interactionContract, context, interaction, providedSource, publication, dataRequirements)»
					«ENDIF»
            	«ENDFOR»
«««            	«val providedContext = (context.behavior.contract as PushInteractionContract).providedData.dataSource.context»
«««            	«val providedSource = (context.behavior.contract as PushInteractionContract).providedData.dataSource.sourceRef»
«««            	«val dataRequirements = (context.behavior.contract as PushInteractionContract).pullInteractionContract != null»
«««            	«val publication = context.behavior.behaviorPublication»
«««            	«IF providedContext != null»
«««            	«valueReceivedFromContext(context, providedContext, publication, dataRequirements)»
«««            	«ELSE»
«««            	«valueReceivedFromDevice(context, providedSource, publication, dataRequirements)»
«««            	«ENDIF»               
            }
            

        
        
        @Override
        public final Object getValueCalled(java.util.Map<String, Object> properties, RemoteServiceInfo source, String valueName,
                Object... indexes) throws Exception {
            if (valueName.equals("«context.name»")) {
                «IF context.behaviors.exists[interaction.equals(IConstant.REQUIRED_INTERACTION)]»
                    «context.onGetPublish(context.behaviors.findFirst[interaction.equals(IConstant.REQUIRED_INTERACTION)].contract != null)»
                «ELSE»
                    return getLastValue(«context.indices.parametersInOrder("indexes")»);
                «ENDIF»
            }
            throw new InvocationException("Unsupported method call: " + valueName);
        }
        // End of methods from the Service class
    '''

    def valueReceivedPublish(ContextDef context, BehaviorPublication publishType, String methodCall) {
        return if(publishType == BehaviorPublication::MAYBE) {
            val indices = if(context.indices.nullOrEmpty) "" else ", " + context.indices.map['''returnValue.get«name.toFirstUpper»()'''].join(",\n        ");
            '''
            	«context.name»Value returnValue = «methodCall»;
            	if(returnValue != null && returnValue.doPublish())
            		set«context.name»(returnValue.getValue()«indices»);
            '''
        } else if(publishType == BehaviorPublication::NO) {
            methodCall + ";\n"
        } else if(context.indices.nullOrEmpty) {
            '''
            set«context.name»(«methodCall»);
            '''
        } else {
            '''
            «context.name»Value returnValue = «methodCall»;
            	if(returnValue != null) {
                	set«context.name»(returnValue.value(), «context.indices.map['''returnValue.indices().«name»()'''].join(",\n        ")»);
            	} else {
                	set«context.name»(null, «context.indices.map(["null"]).join(", ")»);
            	}
            '''
        }
    }
       
    def onGetPublish(ContextDef context, boolean hasDiscover) {
        val indices = context.indices.parametersInOrder("indexes");
        val args = if(hasDiscover) (if(indices.empty) "new Discover()" else (indices + ",\n        new Discover()")) else indices;
        return
          	'''
             return getValue(«args»);
			'''
    }
       
    def valueReceivedFromContext(InteractionContractDef interactionContract, ContextDef context, String interaction, ContextDef contextRef, BehaviorPublication publishType, boolean dataRequirements) {
    val periodicDelivery = if(interactionContract.interaction.equals(IConstant.PERIODIC_INTERACTION)) true else false;
    val callbackPrefix = if(periodicDelivery) "onPeriodic" else "on"; 
    
    '''
        if(eventName.equals("«contextRef.name»") && source.isCompatible("«contextRef.classNameSpace»")) {
            «contextRef.name»Value «contextRef.name.toFirstLower»Value = new «contextRef.name»Value((«contextRef.dataTypeRef.fullname») value«
                if(contextRef.indices.nullOrEmpty) ""
                else ",\n        " + contextRef.indices.parametersInOrder("indexes")
            »);
            
            «context.valueReceivedPublish(publishType, '''«callbackPrefix»«contextRef.name»(«contextRef.name.toFirstLower»Value«
                    IF dataRequirements», new Discover()«ENDIF»)'''
            )»
        }
    '''
    }
    
    def valueReceivedFromDevice(InteractionContractDef interactionContract, ContextDef context, String interaction, SourceRef sourceRef, BehaviorPublication publishType, Boolean dataRequirements) '''
	if (eventName.equals("«sourceRef.source.name»") && source.isCompatible("«sourceRef.device.className»")) {
		«IF (interaction.equals(IConstant.PERIODIC_INTERACTION) || (interactionContract.contract as PushInteractionContract).providedData.periodicity != null)»
«««		«IF ((interactionContract.contract as PushInteractionContract).providedData.map || (interactionContract.contract as PushInteractionContract).providedData.reduce)»
			«generateSourceMap(interactionContract, sourceRef)»
    	«ELSE»
    		«sourceRef.sourceFromDevice» «sourceRef.sourceFromDevice.toFirstLower» = new «sourceRef.sourceFromDevice»(this, source, («sourceRef.source.dataTypeRef.fullname») value«
                if(sourceRef.source.indices.nullOrEmpty) ""
                else ",\n        " + sourceRef.source.indices.parametersInOrder("indexes")
            »);
		«ENDIF»
		
		«IF ((interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION)) && ((interactionContract.contract as PushInteractionContract).providedData.map || (interactionContract.contract as PushInteractionContract).providedData.reduce))»
			«context.valueReceivedPublish(publishType, '''on«sourceRef.sourceFromDevice»(«sourceRef.source.name»Map«IF dataRequirements», new Discover()«ENDIF»)''')»
		«ELSEIF (interaction.equals(IConstant.EVENT_DRIVEN_INTERACTION))»
			«context.valueReceivedPublish(publishType, '''on«sourceRef.sourceFromDevice»(«sourceRef.sourceFromDevice.toFirstLower»«IF dataRequirements», new Discover()«ENDIF»)''')»
		«ELSE»
			«context.valueReceivedPublish(publishType, '''onPeriodic«sourceRef.sourceFromDevice»(«sourceRef.source.name»Map«IF dataRequirements», new Discover()«ENDIF»)''')»
		«ENDIF»
	}
	'''
    
    def generateSourceMap(InteractionContractDef interactionContract, SourceRef sourceRef) {
    	val sourceTypeName = if(sourceRef.source.dataTypeRef.isCustomType)
    						sourceRef.source.dataTypeRef.type.name
    					else 
    						sourceRef.source.dataTypeRef.primitiveTypeRef.literal   	
    	
    	val deliveryPeriodicity = (interactionContract.contract as PushInteractionContract).providedData.dataSource.periodicity;					
//    	val deliveryPeriodicity = sourceRef.periodicity;
		val groupingPeriodicity = (interactionContract.contract as PushInteractionContract).providedData.periodicity;    						
    	val valuesNumber = if(groupingPeriodicity != null && deliveryPeriodicity != null) (groupingPeriodicity.periodicityInSeconds/deliveryPeriodicity.periodicityInSeconds) else 1;
    	    	
    	if((interactionContract.contract as PushInteractionContract).hasGroupingAttribute) {
			val groupingAttribute = (interactionContract.contract as PushInteractionContract).providedData.groupingAttribute		
			val groupingAttributeTypeName = if(groupingAttribute.attrTypeRef.isCustomType) 
											groupingAttribute.attrTypeRef.type.name 
										else 
											groupingAttribute.attrTypeRef.primitiveTypeRef.literal 
//			val customAttrType = groupingAttribute.dataTypeRef.type
//			val primitiveAttrType = groupingAttribute.dataTypeRef.primitiveTypeRef		
			
//			var String groupingAttributeType = null
//			var String sourceType = null
//		
//			groupingAttributeType = if(customAttrType != null) customAttrType.name else primitiveAttrType.literal			
    		val dataRequirement = (interactionContract.contract as PushInteractionContract).providedData
    		
    		if(dataRequirement.map && dataRequirement.reduce) {
    		
    		return '''
    		Map<«groupingAttributeTypeName», «dataRequirement.reduceType.typeName.split('\\.').last»> «sourceRef.source.name»Map = Map.class.cast(value);'''
//    		Map<«groupingAttributeTypeName», «dataRequirement.reduceType.literal.split('\\.').last»> «sourceRef.source.name»Map = Map.class.cast(value);'''    			
    		
    		} else if(dataRequirement.map) { 
    		return '''
    		Map<«groupingAttributeTypeName», List<«dataRequirement.mapOutputValueType.typeName.split('\\.').last»>> «sourceRef.source.name»Map = Map.class.cast(value);'''
    		
    		} else {    		
    		return '''
    		Map<«groupingAttributeTypeName», 
    		«IF valuesNumber == 1»
    		List<«sourceTypeName»>> «sourceRef.source.name»Map
    		«ELSE»
    		List<Buffer<«sourceTypeName»>>> «sourceRef.source.name»Map
    		«ENDIF»
    		 = Map.class.
    				cast(DeviceSourceBuffer.class.cast(value).
    						getContentGroupedByAttribute("«groupingAttribute.name»", «valuesNumber»));'''
    	
    		}
		}
		
		else
			return '''
    		Map<String, List<«sourceTypeName»>> «sourceRef.source.name»Map = Map.class.
    				cast(DeviceSourceBuffer.class.cast(value).
    						getContentGroupedByAttribute("id", 1));'''
    }
    
    /* Create the discoverer class for one behavior */
    def discoverCodeForBehavior(InteractionContractDef interactionContract) {
        val name = if(interactionContract.interaction.equals(IConstant.REQUIRED_INTERACTION)) "Discovery" else
            if((interactionContract.contract as PushInteractionContract).providedData.dataSource.context != null)
            	(interactionContract.contract as PushInteractionContract).providedData.dataSource.context.name
            else
            	(interactionContract.contract as PushInteractionContract).providedData.dataSource.sourceRef.source.name       	
//        val comment = behaviorDef.stripComments;
        
        val contractName = if(interactionContract.interaction.equals(IConstant.REQUIRED_INTERACTION)) "when required" else
            if((interactionContract.contract as PushInteractionContract).providedData.dataSource.context != null) 
            	"when provided " + (interactionContract.contract as PushInteractionContract).providedData.dataSource.context.name
            else "when provided " + (interactionContract.contract as PushInteractionContract).providedData.dataSource.sourceRef.source.name
        
        val shortDesc = if(interactionContract.interaction.equals(IConstant.REQUIRED_INTERACTION)) "get" else name
		
		var Iterable<SourceRef> sources = new ArrayList<SourceRef>();
		var Iterable<ContextDef> contexts = new ArrayList<ContextDef>();
		
		println("Handling " + interactionContract.interaction + " " + interactionContract.contract.toString)
		
		if(interactionContract.interaction.equals(IConstant.REQUIRED_INTERACTION)) {
				sources =  (interactionContract.contract as PullInteractionContract).dataRequirements.filter[req | req.dataSource.sourceRef !=null].map[req | req.dataSource.sourceRef]
				contexts = (interactionContract.contract as PullInteractionContract).dataRequirements.filter[req | req.dataSource.context !=null].map[req | req.dataSource.context]			
		} else {
				sources =  (interactionContract.contract as PushInteractionContract).pullInteractionContract.dataRequirements.filter[req | req.dataSource.sourceRef != null].map[req | req.dataSource.sourceRef]
				contexts = (interactionContract.contract as PushInteractionContract).pullInteractionContract.dataRequirements.filter[req | req.dataSource.context != null].map[req | req.dataSource.context]
		}
        
        return interactionContract.contextContainer.discoverCodeForBehavior(name.split(" ").map[toFirstUpper].join,
                            /*comment*/"", contractName, shortDesc,        
                           contexts,
                           sources)
    }
    
//    def static ContextDef contextContainer(EObject elt) {
//        return (getParentOfClass(elt, ContextDef) as ContextDef);
//    }
//    
//    def static EObject getParentOfClass(EObject elt, Class<? extends EObject> clazz) {
//        if (elt instanceof DomainModel) {
//            return null;
//        }
//        if (clazz.isInstance(elt)) {
//            return elt;
//        } else {
//            return getParentOfClass(elt.eContainer(), clazz);
//        }
//    }
    
    
    /* Create the discoverer class for one behavior */
    def discoverCodeForBehavior(ContextDef context, String name, String comment, String contractName, String shortDesc, Iterable<ContextDef> contexts, Iterable<SourceRef> sources) {
        val devices = sources.map[device].toSet;
        return '''
        // Discover object for «shortDesc»
        «discovererClassForBehavior(name, "Abstract" + context.name, comment, devices, contexts)»
        «FOR dev : devices»
        
            «dev.discovererClassForDevice("For" + name, '''
            Discover object that will exposes the <code>«dev.name»</code> devices to get their sources for the
            <code>«contractName»</code> interaction contract.''')»
            
            «dev.compositeClassForDevice("For" + name, null, null, '''
            A composite of several <code>«dev.name»</code> devices to get their sources for the
            <code>«contractName»</code> interaction contract.''')»
            
            «dev.proxyClassForDevice("For" + name, null, sources.filter[device == dev].map[source], null, '''
            A proxy to one <code>«dev.name»</code> device to get its sources for the
            <code>«contractName»</code> interaction contract.''')»
        «ENDFOR»
        // End of discover object for «shortDesc»
        '''
    }
    
    def contextValue(ContextDef context) {
        return context.name.valueClass(context.name.toFirstUpper+"Value", 
            context.packageName,
            '''the context <code>«context.name»</code>''',
            context, 
            '''An object to store a value published by the context <code>«context.name»</code>.''',
//            + (if(!context.comments.nullOrEmpty) ("\n<p>\n" + context.comments) else ""),
            context.dataTypeRef,
            context.indices,
            "", "", "", "", ""/*context.comments*/);
    }
}
