package fr.inria.phoenix.diasuite.diaswarm.generator.template.controller

import fr.inria.phoenix.diasuite.diaswarm.diaswarm.ContextDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.ControllerBehaviorDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.ControllerDef
import fr.inria.phoenix.diasuite.diaswarm.generator.template.IConstant
import java.util.HashSet
import java.util.Set

class DiaSwarmControllerHelper {

	/**============================ HELPER METHODS ==============================**/
	def static generateFileName(ControllerDef controller) {
		controller.getPackageName.replaceAll("\\.", "/") + "/" + IConstant.ABSTRACT_CLASS_PREFIX + controller.name +
			IConstant.JAVA_EXTENSION_POSTFIX
	}

	def static getPackageName(ControllerDef controller) {
		IConstant.BASE_PACKAGE_NAMESPACE + "." + IConstant.CONTROLLER_PACKAGE_NAME + "." +
			controller.getName().toLowerCase()
	}

	def static generateFileNameForContextFeature(ControllerDef controller, String controllerFeature) {
		controller.getPackageName.replaceAll("\\.", "/") + "/" + controller.name.toFirstUpper + controllerFeature +
			IConstant.JAVA_EXTENSION_POSTFIX
	}

	/**
	 * Return the full list of contexts (including those defined in interaction contracts) used by a controller
	 * 
	 * @param controller
	 *            The controller to extract contexts from
	 * @return The full list of contexts used by the controller
	 */
	def static Set<ContextDef> getAllContexts(ControllerDef controller) {
		val Set<String> processed = new HashSet<String>();
		val Set<ContextDef> contexts = new HashSet<ContextDef>();
		for (ControllerBehaviorDef behavior : controller.getBehaviors()) {
			val ContextDef cd = behavior.getActivationCondition();
			if (cd != null) {
				if (!processed.contains(cd.getName())) {
					processed.add(cd.getName());
					contexts.add(cd);
				}
			}

			for (ContextDef cd2 : behavior.getDataRequirements()) {
				if (!processed.contains(cd2)) {
					processed.add(cd2.getName());
					contexts.add(cd2);
				}
			}
		}
		return contexts;
	}

	def static Iterable<ContextDef> providedContexts(ControllerDef controller) {
		return controller.behaviors.map([activationCondition]);
	}

}
