package fr.inria.phoenix.diasuite.diaswarm.generator.template.datatype

import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DataTypeRef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.EnumDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.PrimitiveTypeRef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.SourceType
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.StructDef
import fr.inria.phoenix.diasuite.diaswarm.generator.template.IConstant

class DiaSwarmDatatypeHelper {

	/**============================ HELPER METHODS ==============================**/
	/**
	 * Returns a string representing the filename of an enumeration java class 
	 * along with it's namespace.
	 */
	def static generateFileName(EnumDef enumeration) {
		enumeration.packageName.replaceAll("\\.", "/") + "/" + enumeration.name + ".java"
	}

	def static generateFileName(EnumDef enumeration, String postFix) {
		enumeration.packageName.replaceAll("\\.", "/") + "/" + enumeration.name + postFix + ".java"
	}

	/**
	 * Returns a string representing the filename of a structure java class 
	 * along with it's namespace.
	 */
	def static generateFileName(StructDef structure) {
		structure.packageName.replaceAll("\\.", "/") + "/" + structure.name + ".java"
	}

	def static generateFileName(StructDef structure, String postFix) {
		structure.packageName.replaceAll("\\.", "/") + "/" + structure.name + postFix + ".java"
	}

	/**
	 * Returns a string representing the package namespace for structures.  
	 */
	def static getPackageName(StructDef structure) {
		IConstant.BASE_PACKAGE_NAMESPACE + "." + IConstant.DATATYPE_PACKAGE_NAME + "." + structure.name.toLowerCase
	}

	/**
	 * Returns a string representing the package namespace for sources.
	 */
	def static getPackageName(SourceType source) {
		IConstant.BASE_PACKAGE_NAMESPACE + "." + IConstant.DATATYPE_PACKAGE_NAME + "." + source.name.toLowerCase
	}

	/**
	 * Returns a string representing the data-type of the field 
	 * provided by parameter.
	 */
	def static String getFullName(DataTypeRef dataType) {
		if (dataType.list) '''java.util.List<«dataType.typeName»>''' else
			dataType.typeName
	}

	/**
	 * Returns a string representing the data-type of the field 
	 * provided by parameter.
	 */
	def static getTypeName(DataTypeRef dataType) {
		if (dataType.type != null) {
			dataType.type.packageName + "." + dataType.type.name
		} else {
			switch dataType.primitiveTypeRef {
				case PrimitiveTypeRef::BOOLEAN: IConstant.BOOLEAN
				case PrimitiveTypeRef::INTEGER: IConstant.INTEGER
				case PrimitiveTypeRef::FLOAT: IConstant.FLOAT
				case PrimitiveTypeRef::STRING: IConstant.STRING
				case PrimitiveTypeRef::BINARY: IConstant.BINARY
			}
		}
	}

/*============================================================================*/
}
