package fr.inria.phoenix.diasuite.diaswarm.generator.template.deployment

import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DataTypeRef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DomainModel
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.PushInteractionContract
import java.util.Map
import java.util.Map.Entry
import java.util.Set

import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.context.DiaSwarmContextHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.controller.DiaSwarmControllerHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.deployment.DiaSwarmDeploymentHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.datatype.DiaSwarmDatatypeHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.device.DiaSwarmDeviceHelper.*

class DiaSwarmDeployment {
	
	def generateDeploymentCode(Iterable<DomainModel> models, Set<String> deviceSourceBufferSet) {
		mainDeploy(models, packageName, deviceSourceBufferSet)
	} 
	
	def generateBufferForDeviceSource(Iterable<DomainModel> models, String bufferName, String sourceName, DataTypeRef sourceType, Integer deviceSourceBufferSize, 
										Map<String, String> deviceSourceToGroupingAttributes)'''
		«var String sourceTypeName = sourceType.fullName»
		package fr.inria.phoenix.diasuite.framework.deployment;
		
		import java.io.BufferedWriter;
		import java.io.FileWriter;
		import java.io.IOException;
		import java.util.List;
		import java.util.ArrayList;
		import java.util.Set;
		import java.util.HashSet;
		import java.util.Map;
		import java.util.Map.Entry;
		import java.util.HashMap;
		
		import fr.inria.diagen.core.service.buffer.Buffer;
		import fr.inria.diagen.core.service.buffer.CircularFifoBuffer;
		import fr.inria.diagen.core.service.buffer.DeviceSourceBuffer;
		
		import «sourceTypeName»;
		«IF deviceSourceToGroupingAttributes != null»
			«FOR attributeName : deviceSourceToGroupingAttributes.keySet»
				«var attributeType = deviceSourceToGroupingAttributes.get(attributeName)»
				«var String[] typeSplit = attributeType.split('\\.')»
				«IF typeSplit.size > 1»
					import «attributeType»;
«««					«var oldType = deviceSourceToGroupingAttributes.put(attributeName, typeSplit.last)»
				«ENDIF»			
			«ENDFOR»
		«ENDIF»
		
		public class «bufferName.toFirstUpper» implements DeviceSourceBuffer {
			
			private Map<String, Buffer<«sourceTypeName.split('\\.').last»>> idTo«sourceName.toFirstUpper»Map;
			private Map<String, Map<String, Object>> idToAttributes;
			«IF deviceSourceToGroupingAttributes != null»
				«FOR attributeName : deviceSourceToGroupingAttributes.keySet»
					«IF !attributeName.equals("id")»
						private Map<«deviceSourceToGroupingAttributes.get(attributeName)», Set<String>> «attributeName»ToIdsMap;	
					«ENDIF»
				«ENDFOR»
			«ENDIF»
			private static final int CAPACITY = «deviceSourceBufferSize»;
			
			public «bufferName.toFirstUpper»() {
				idTo«sourceName.toFirstUpper»Map = new HashMap<String, Buffer<«sourceTypeName.split('\\.').last»>>();
				idToAttributes = new HashMap<String, Map<String, Object>>();
				«IF deviceSourceToGroupingAttributes != null»
					«FOR attributeName : deviceSourceToGroupingAttributes.keySet»
						«IF !attributeName.equals("id")»				
							«attributeName»ToIdsMap = new HashMap<«deviceSourceToGroupingAttributes.get(attributeName)», Set<String>>();
						«ENDIF»
					«ENDFOR»
				«ENDIF»
			}
			
			@Override
			public void add(Map<String, Object> attributes, Object sourceValue) {
				
				String id = (String) attributes.get("id");
				«IF deviceSourceToGroupingAttributes != null»
					«FOR attributeName : deviceSourceToGroupingAttributes.keySet»
						«IF !attributeName.equals("id")»		
						«deviceSourceToGroupingAttributes.get(attributeName)» «attributeName» = («deviceSourceToGroupingAttributes.get(attributeName)») attributes.get("«attributeName»");
						«ENDIF»
					«ENDFOR»
				«ENDIF»
				
				Set<String> idSet = new HashSet<String>();
				«IF deviceSourceToGroupingAttributes != null»
				if (id != null «FOR attributeName : deviceSourceToGroupingAttributes.keySet» && «attributeName» != null«ENDFOR») {
					Buffer<«sourceTypeName.split('\\.').last»> sourceBuffer;
				«ELSE»
				if (id != null) {
					Buffer<«sourceTypeName.split('\\.').last»> sourceBuffer;
				«ENDIF»
				
				
					if ((sourceBuffer = idTo«sourceName.toFirstUpper»Map.get(id)) == null) {
						sourceBuffer = new CircularFifoBuffer<«sourceTypeName.split('\\.').last»>(
							CAPACITY);
						idTo«sourceName.toFirstUpper»Map.put(id, sourceBuffer);
					}
					sourceBuffer.add((«sourceTypeName.split('\\.').last») sourceValue);
					
					Map<String, Object> attributeMap;
					if ((attributeMap = idToAttributes.get(id)) == null) {
						attributeMap = new HashMap<String, Object>();
						idToAttributes.put(id, attributeMap);
					}
					attributeMap.putAll(attributes);
					
					«IF deviceSourceToGroupingAttributes != null»
						«FOR attributeName : deviceSourceToGroupingAttributes.keySet»
							«IF !attributeName.equals("id")»
						if ((idSet = «attributeName»ToIdsMap.get(«attributeName»)) == null) {
							idSet = new HashSet<String>();
							«attributeName»ToIdsMap.put(«attributeName», idSet);
						}
							«ENDIF»
						«ENDFOR»
					«ENDIF»
					idSet.add(id);
				
				} else
					throw new NullPointerException(
								"The provided attributes can't be null !");
			}
			
			public String toString() {
				StringBuffer buffer = new StringBuffer();

				for (String id : idTo«sourceName.toFirstUpper»Map.keySet()) {
					buffer.append(id + ": [");
					buffer.append(idTo«sourceName.toFirstUpper»Map.get(id).toString() + "]");
					buffer.append("\n");
				}

			return buffer.toString();
			}
			
			@Override
			public int size() {
				// TODO Auto-generated method stub
				return 0;
			}
		
			@Override
			public Object getContent() {
				return getContentGroupedByAttribute("id", CAPACITY);
			}
		
			@Override
			public Object getContentGroupedByAttribute(String attribute, int size) {
				switch (attribute) {
				«IF deviceSourceToGroupingAttributes != null»
					«FOR Entry<String,String> entry : deviceSourceToGroupingAttributes.entrySet»
					«IF !entry.key.equals("id")»	
					case "«entry.key»": {
					if(size == 1) {
						Map<«entry.value», List<«sourceTypeName.split('\\.').last»>> «sourceName»By«entry.key.toFirstUpper»AttributeMap = new HashMap<«entry.value», List<«sourceTypeName.split('\\.').last»>>();
						List<«sourceTypeName.split('\\.').last»> «sourceName»ListFor«entry.key.toFirstUpper»;
						
						for (Entry<«entry.value», Set<String>> «entry.key»Devices : «entry.key»ToIdsMap
							.entrySet()) {
							«sourceName»ListFor«entry.key.toFirstUpper» = new ArrayList<«sourceTypeName.split('\\.').last»>();
							
							for (String deviceId : «entry.key»Devices.getValue()) {
								«sourceName»ListFor«entry.key.toFirstUpper».addAll(idTo«sourceName.toFirstUpper»Map.get(deviceId)
										.getLastElement().toList());
							}
							«sourceName»By«entry.key.toFirstUpper»AttributeMap.put(«entry.key»Devices.getKey(), «sourceName»ListFor«entry.key.toFirstUpper»);
						}
						
						return «sourceName»By«entry.key.toFirstUpper»AttributeMap;
					} else {
						Map<«entry.value», List<Buffer<«sourceTypeName.split('\\.').last»>>> «sourceName»By«entry.key.toFirstUpper»AttributeMap = new HashMap<«entry.value», List<Buffer<«sourceTypeName.split('\\.').last»>>>();
						List<Buffer<«sourceTypeName.split('\\.').last»>> «sourceName»ListFor«entry.key.toFirstUpper» = new ArrayList<Buffer<«sourceTypeName.split('\\.').last»>>();
						
						for (Entry<«entry.value», Set<String>> «entry.key»Devices : «entry.key»ToIdsMap
							.entrySet()) {
							«sourceName»ListFor«entry.key.toFirstUpper» = new ArrayList<Buffer<«sourceTypeName.split('\\.').last»>>();
							
							for (String deviceId : «entry.key»Devices.getValue()) {
								«sourceName»ListFor«entry.key.toFirstUpper».add(idTo«sourceName.toFirstUpper»Map
								.get(deviceId).getLastElements(size));
«««								«sourceName»ListFor«entry.key.toFirstUpper» = idTo«sourceName.toFirstUpper»Map.get(deviceId)
«««										.toList(size);
							}
							«sourceName»By«entry.key.toFirstUpper»AttributeMap.put(«entry.key»Devices.getKey(), «sourceName»ListFor«entry.key.toFirstUpper»);
						}
						
						return «sourceName»By«entry.key.toFirstUpper»AttributeMap;
					}
					
					}
					«ENDIF»
					«ENDFOR»
				«ENDIF»
				
				case "id": {
				if (size == 1) {
					Map<String, List<«sourceTypeName.split('\\.').last»>> «sourceName»ByIdAttributeMap = new HashMap<String, List<«sourceTypeName.split('\\.').last»>>();
					List<«sourceTypeName.split('\\.').last»> «sourceName»ListForId = new ArrayList<«sourceTypeName.split('\\.').last»>();
					for (Entry<String, Buffer<«sourceTypeName.split('\\.').last»>> device«sourceName.toFirstUpper» : idTo«sourceName.toFirstUpper»Map
						.entrySet()) {
						«sourceName»ListForId = device«sourceName.toFirstUpper».getValue().getLastElement().toList();
						«sourceName»ByIdAttributeMap.put(device«sourceName.toFirstUpper».getKey(), «sourceName»ListForId);
					}
					
					return «sourceName»ByIdAttributeMap;
					
				} else {
					Map<String, Buffer<«sourceTypeName.split('\\.').last»>> «sourceName»ByIdAttributeMap = new HashMap<String, Buffer<«sourceTypeName.split('\\.').last»>>();
«««					List<«sourceType.split('\\.').last»> «sourceName»ListForId = new ArrayList<«sourceType.split('\\.').last»>();
					for (Entry<String, Buffer<«sourceTypeName.split('\\.').last»>> device«sourceName.toFirstUpper» : idTo«sourceName.toFirstUpper»Map
						.entrySet()) {
«««						«sourceName»ListForId = device«sourceName.toFirstUpper».getValue().toList();
						«sourceName»ByIdAttributeMap.put(device«sourceName.toFirstUpper».getKey(), 
								device«sourceName.toFirstUpper».getValue().getLastElements(size));
					}
					
					return «sourceName»ByIdAttributeMap;
					
					}
				}
		
				default:
					try {
						throw new Exception(attribute + " is not a valid grouping attribute !");
					} catch (Exception e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
				
				return null;
			}
		
			@Override
			public void saveContentsToFile(String fileName, int size) {
				try {
					FileWriter fileWritter = new FileWriter(fileName, true);
					BufferedWriter bufferWritter = new BufferedWriter(fileWritter);
					
					StringBuilder builder = new StringBuilder();
					
					for (Entry<String, Buffer<«sourceTypeName.split('\\.').last»>> «sourceName»Buffer : idTo«sourceName.toFirstUpper»Map
							.entrySet()) {
						int i = 1;
						for («sourceTypeName.split('\\.').last» «sourceName» : «sourceName»Buffer.getValue().getLastElements(size).toList()) {
							builder.append("{");
							builder.append("\"«sourceName»\" : "
							«IF sourceType.type != null»
							+ «sourceName».toJSON());
							«ELSE»
							+ "\"" + «sourceName».toString() + "\"");
							«ENDIF»

						for (Entry<String, Object> attributeEntry : idToAttributes
							.get(«sourceName»Buffer.getKey()).entrySet())
							builder.append(", \"" + attributeEntry.getKey()
								+ "\": " + "\"" + attributeEntry.getValue()
								+ "\"");

						builder.append(", \"position\" : " + "\"" + i++ + "\"");
						builder.append("}");
						builder.append("\n");
						}
					}

					bufferWritter.write(builder.toString());

					builder.setLength(0);
					bufferWritter.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
			
			@Override
			public int capacity() {
				return CAPACITY;
			}
			
			@Override
			public void clear() {
				for (Entry<String, Buffer<«sourceTypeName.split('\\.').last»>> entry : idTo«sourceName.toFirstUpper»Map.entrySet()) {
						entry.getValue().clear();
					}
				}
		}
	'''	
	
    def mainDeploy(Iterable<DomainModel> models, String deploymentPackageName, Set<String> deviceSourceBufferSet) '''
        package «deploymentPackageName»;
        
        import fr.inria.diagen.core.deploy.AbstractDeploy;
        
        «models.contexts.map[context | "import " + context.packageName + ".Abstract" + context.name.toFirstUpper + ";"].join("\n")»
        
        «IF deviceSourceBufferSet.size != 0»
        import fr.inria.diagen.core.service.buffer.DeviceSourceBuffer;
        «ENDIF»
        
        «var mrImports = 0»
        «FOR context : models.contexts»
        	«IF context.mapReduceBehaviors.size != 0»
        		import «context.getPackageName».«context.name.toFirstUpper»Job;
        		import «context.getPackageName».«context.name.toFirstUpper»ResultsParser;
        		«IF (mrImports = mrImports + 1).equals(1)» 
        		import org.apache.hadoop.util.Tool;
        		import fr.inria.diagen.core.network.TimeUnit;
        		import fr.inria.diagen.core.mapreduce.MapReduceConfiguration;
        		import fr.inria.diagen.core.mapreduce.ProcessingStagesEnum;
        		import fr.inria.diagen.core.service.local.ServiceProperties;
        		«ENDIF»
        	«ENDIF»
        «ENDFOR»

        «models.controllers.map[controller | "import " + controller.packageName + ".Abstract" + controller.name.toFirstUpper + ";"].join("\n")»
        
        /**
         * This class should be implemented to bind the implementation of the various components
         */
        public abstract class AppComponentBinder extends AbstractDeploy {
        
            // Context instances
            «models.contexts.map["private Abstract" + name.toFirstUpper + " " + name.toFirstLower + "Instance = null;"].join("\n")»
        
            // Controller instances
            «models.controllers.map["private Abstract" + name.toFirstUpper + " " + name.toFirstLower + "Instance = null;"].join("\n")»
            
            @Override
            public void deployAll() {
            	
            	«FOR bufferName : deviceSourceBufferSet»
            		«var sourceName = bufferName.split("BufferFor").get(0)»
            		«var deviceName = bufferName.split("BufferFor").get(1)»
            	initializeSourceBuffer("«sourceName»", 
            							get«sourceName.toFirstUpper»From«deviceName.toFirstUpper»BufferClass());
            	«ENDFOR»
                // Initialization of contexts
                «models.contexts.map['''
                    if («name.toFirstLower»Instance == null) {
                        «name.toFirstLower»Instance = getServiceInstance(get«name.toFirstUpper»Class());
                    	«FOR behavior : mapReduceBehaviors»
                    		set«name.toFirstUpper»MapReduceConfig();
                    	«ENDFOR»
                    }    
                '''].join»
                // Intialization of controllers
                «models.controllers.map['''
                    if («name.toFirstLower»Instance == null)
                        «name.toFirstLower»Instance = getServiceInstance(get«name.toFirstUpper»Class());
                '''].join»
                // Deploying contexts
                «models.contexts.map["deploy(" + name.toFirstLower + "Instance);"].join("\n")»
                // Deploying controllers
                «models.controllers.map["deploy(" + name.toFirstLower + "Instance);"].join("\n")»
            }
            
            @Override
            public void undeployAll() {
                // Undeploying contexts
                «models.contexts.map["undeploy(" + name.toFirstLower + "Instance);"].join("\n")»
                // Undeploying controllers
                «models.controllers.map["undeploy(" + name.toFirstLower + "Instance);"].join("\n")»
            }
            // Abstract binding methods for contexts
            
            «FOR bufferName : deviceSourceBufferSet»
            	«var sourceName = bufferName.split("BufferFor").get(0).toFirstUpper»
            	«var deviceName = bufferName.split("BufferFor").get(1).toFirstUpper»
            	public Class<? extends DeviceSourceBuffer> get«sourceName»From«deviceName»BufferClass() {
            		return «sourceName»BufferFor«deviceName».class;
            	}
            «ENDFOR»
            
            «FOR context : models.contexts»
            	«FOR behavior : context.mapReduceBehaviors»
«««            	«val sourceName = (behavior.contract as PushInteractionContract).providedData.dataSource.sourceRef.device.name»
            	«val contextName = context.name»
            	«val period = (behavior.contract as PushInteractionContract).providedData.dataSource.periodicity»
            	«val groupingPeriod = (behavior.contract as PushInteractionContract).providedData.periodicity»
            	
            	«val mapDefined = (behavior.contract as PushInteractionContract).providedData.map»
            	«val reduceDefined = (behavior.contract as PushInteractionContract).providedData.reduce»
            	
            	public Class<? extends Tool> get«contextName»JobClass() {
            		return «contextName»Job.class;
            	}
            	
            	«IF groupingPeriod != null»
            	public void set«contextName.toFirstUpper»MapReduceConfig() {
            		MapReduceConfiguration «contextName»MapReduce = new MapReduceConfiguration.Builder()
            				.processingStages(«if(mapDefined && reduceDefined) "ProcessingStagesEnum.MAPREDUCE" else "ProcessingStagesEnum.MAP"»)
            				.contexJobClass(get«contextName»JobClass())
            				.resultsFunc(new «contextName.toFirstUpper»ResultsParser())
            				.contexClass(get«contextName»Class()).period(«groupingPeriod.value»)
            				.timeUnit(TimeUnit.«groupingPeriod.timeUnit.getName()»).build();
            «ELSE»
            	«val timeUnit = (behavior.contract as PushInteractionContract).providedData.dataSource.periodicity.timeUnit.getName()»
            	public void set«contextName.toFirstUpper»MapReduceConfig() {
            		MapReduceConfiguration «contextName»MapReduce = new MapReduceConfiguration.Builder()
            				.processingStages(«if(mapDefined && reduceDefined) "ProcessingStagesEnum.MAPREDUCE" else "ProcessingStagesEnum.MAP"»)
            				.contexJobClass(get«contextName»JobClass())
            				.resultsFunc(new «contextName.toFirstUpper»ResultsParser())
            				.contexClass(get«contextName»Class()).period(«period.value»)
            				.timeUnit(TimeUnit.«timeUnit»).build();
            «ENDIF»
            
            		«contextName.toFirstLower»Instance.addToOutProperties(
            			ServiceProperties.MAPREDUCE.toString(),
            			«contextName»MapReduce);
            	}
            	«ENDFOR»
            «ENDFOR»
            «FOR c : models.contexts»
                public abstract Class<? extends Abstract«c.name.toFirstUpper»> get«c.name.toFirstUpper»Class();
                
            «ENDFOR»
            // End of abstract binding methods for contexts
            
            // Abstract binding methods for controllers
            «FOR c : models.controllers»
«««                «c.javadoc("Overrides this method to provide the implementation class of the <code>" + c.name + "</code> controller",
«««                    "\n@return a class object of a derivation of {@link Abstract" + c.name.toFirstUpper + "} that implements the <code>" + c.name + "</code> controller",
«««                    0
«««                )»
                public abstract Class<? extends Abstract«c.name.toFirstUpper»> get«c.name.toFirstUpper»Class();
                
            «ENDFOR»
            // End of abstract binding methods for controllers
        }
    '''
    
    def deviceList(DomainModel model, String packageName) '''
        package «packageName»;
        
        import java.util.HashMap;
        import java.util.Map;
        
        import fr.inria.diagen.core.service.local.Service;
        
        /**
         * A container for the list of available devices. For internal use by introspection libraries
         * 
         * @internal
         */
        public class ClassList implements fr.inria.diagen.core.deploy.DeviceLister {
        
            public final Map<String, Class<? extends Service>> deviceList = new HashMap<String, Class<? extends Service>>();
        
            public ClassList() {
                «FOR device : model.devices»
                    deviceList.put("«device.className»", «device.packageName».Abstract«device.name.toFirstUpper».class);
                «ENDFOR»
            }
        
            public Map<String, Class<? extends Service>> getDeviceList() {
                return deviceList;
            }
        
        }
    '''
}