package fr.inria.phoenix.diasuite.diaswarm.generator.template.deployment

import fr.inria.phoenix.diasuite.diaswarm.diaswarm.ContextDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.ControllerDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DeviceType
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DomainModel
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.Periodicity
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.TimeUnit
import fr.inria.phoenix.diasuite.diaswarm.generator.template.IConstant
import java.util.ArrayList
import java.util.HashMap
import java.util.List

class DiaSwarmDeploymentHelper {

	/**============================ HELPER METHODS ==============================**/
	def static generateDeploymentFileName(Iterable<DomainModel> models) {
		getPackageName.replaceAll("\\.", "/") + "/" + IConstant.DEPLOYMENT_FILE_NAME + IConstant.JAVA_EXTENSION_POSTFIX
	}

	def static generateBufferFileName(Iterable<DomainModel> models, String bufferName) {
		getPackageName.replaceAll("\\.", "/") + "/" + bufferName.toFirstUpper + IConstant.JAVA_EXTENSION_POSTFIX
	}

	def static getPackageName() {
		IConstant.BASE_PACKAGE_NAMESPACE + "." + IConstant.DEPLOYMENT_PACKAGE_NAME
	}

	def static <T> Iterable<T> uniq(Iterable<? extends T> original, (T)=>Object func) {
		val values = new HashMap<Object, T>();
		original.forEach(o|values.put(func.apply(o), o));
		return values.values;
	}

	def static Iterable<ContextDef> getContexts(DomainModel model) {
		var List<ContextDef> contexts = new ArrayList<ContextDef>()

		for (element : model.elements)
			if (element instanceof ContextDef)
				contexts.add(element as ContextDef)

		contexts
	}

	def static Iterable<ContextDef> getContexts(Iterable<DomainModel> modelIterable) {
		println("Recovering contexts via getContext method ! : modelIterable size  : " + modelIterable.size)
		var List<ContextDef> contexts = new ArrayList<ContextDef>()

		for (model : modelIterable) {
			println("Handling model : " + model.toString)
			contexts.addAll(model.contexts)
		}

		contexts
	}

	def static Iterable<ControllerDef> getControllers(DomainModel model) {
		var List<ControllerDef> controllers = new ArrayList<ControllerDef>()

		for (element : model.elements)
			if (element instanceof ControllerDef)
				controllers.add(element as ControllerDef)

		controllers
	}

	def static Iterable<ControllerDef> getControllers(Iterable<DomainModel> modelIterable) {
		var List<ControllerDef> controllers = new ArrayList<ControllerDef>()

		for (model : modelIterable)
			controllers.addAll(model.controllers)

		controllers
	}

	def static Iterable<DeviceType> getDevices(DomainModel model) {
		var List<DeviceType> devices = new ArrayList<DeviceType>()

		for (element : model.elements) {
			println("Entering getDevices method. Elements count : " + model.elements.size);
			if (element instanceof DeviceType) {
				println("Elements is of DeviceType : " + element.name);
				devices.add(element as DeviceType)
			}
		}

		devices
	}

	def static Iterable<DeviceType> getDevices(Iterable<DomainModel> modelIterable) {
		println("Recovering devices via getDevices method ! : modelIterable size  : " + modelIterable.size)
		var List<DeviceType> devices = new ArrayList<DeviceType>()

		for (model : modelIterable)
			devices.addAll(model.devices)

		devices
	}

	def static Integer getPeriodicityInSeconds(Periodicity periodicity) {
		switch (periodicity.timeUnit) {
			case TimeUnit.S: return periodicity.value
			case TimeUnit.MIN: return periodicity.value * 60
			case TimeUnit.HR: return (periodicity.value * 60) * 60
			default: return null
		}
	}

}
