package fr.inria.phoenix.diasuite.diaswarm.generator.template.device

import fr.inria.phoenix.diasuite.diaswarm.diaswarm.AttributeDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DeviceType
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.OrderDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.SourceDef
import fr.inria.phoenix.diasuite.diaswarm.generator.template.DiaSwarmEntity
import java.util.List

import static fr.inria.phoenix.diasuite.diaswarm.generator.template.GenericTemplate.*
import static fr.inria.phoenix.diasuite.diaswarm.generator.template.GenericTemplate.*

import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.datatype.DiaSwarmDatatypeHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.device.DiaSwarmDeviceHelper.*

class DiaSwarmDevice extends DiaSwarmEntity {	

/**=============================== TEMPLATES ================================**/

	/* Template for the abstract class for the device */
	/**
	 * 
	 */
	def generateDevice(DeviceType device) {
		println("DEVICE : " + device.name.toString)
		val sources = device.allSources.sortBy[sources | sources.name]
		println("DEVICE SOURCES : " + sources.toString)
	
		val attributes = device.allAttributes.sortBy[attributes | attributes.name]
		println("DEVICE ATTRIBUTES : " + attributes.toString)
		
	
		
		val actions = device.allActionMethods.sortBy[actions | actions.name]
		
		'''
			package «device.packageName»;
			
			«device.generateImports(sources)»
			
			
			public abstract class Abstract«device.name.toFirstUpper» extends Service {
			    
			    /**================= DEVICE ATTRIBUTES =================**/
«««			    private String _id;
			    «FOR attr : attributes»
			    	«attr.generateAttributeFields»
			    «ENDFOR»
			    /*=======================================================*/
			    
			    
				/**=================== DEVICE SOURCES ==================**/
				«FOR source : sources»
					«source.generateSourceFields»
				«ENDFOR»
				
«««				protected void updateId(String newIdValue) {
«««					if (_id != newIdValue) {
«««						_id = newIdValue;
«««						updateAttribute("_id", newIdValue);
«««					}
«««				}
«««				
«««				public String getId() {
«««					return _id;
«««				}		
			    /*=======================================================*/
				
			    
				/**================= DEVICE CONSTRUCTOR ================**/
				
				«device.generateConstructor(attributes)»
				/*=======================================================*/
			    
			    
			    /**=============== METHODS FOR ATTRIBUTES ==============**/
			    «FOR attr : attributes»
			    
			    «attr.generateAttributeMethods»
			    «ENDFOR»
			    /*=======================================================*/
			    
			    
			    /**================ METHODS FOR SOURCES ================**/
			    «FOR source : sources»
			    
			    «source.generateSourceMethods»
			    «ENDFOR»
			    /*=======================================================*/
			    
			    
				/**================ METHODS FOR ACTIONS ================**/
			    «FOR act : actions»
			    	
			    	«act.generateActionMethods()»
			    «ENDFOR»
			    /*=======================================================*/
			    
			    
			    /**================== SERVICE METHODS ==================**/
			    «generateServiceFunctions(attributes, actions, sources)»
			    /*=======================================================*/
			}
		'''
	}
	
	def generateImports(DeviceType device, List<SourceDef> sources) '''		
		import fr.inria.diagen.core.ServiceConfiguration;
		import fr.inria.diagen.core.exception.InvocationException;
		import fr.inria.diagen.core.network.RemoteServiceInfo;
		import fr.inria.diagen.core.network.DataDeliveryModel;
		import fr.inria.diagen.core.service.local.Service;
		import java.util.List;
		
		«IF sources != null && sources.size != 0»
			import «device.packageName».source.*;
		«ENDIF»
	'''

	/* Template for the constructor of the device */
	/**
	 * 
	 */
	def generateConstructor(DeviceType device, List<AttributeDef> attributes) {
		val attributeList = attributes.sortBy[attr | attr.name].map([",\n        " + attrTypeRef.fullName + " " + name + "Value"]).join("");
//									.concat(",\n        String idValue");
		return 
		'''
			public Abstract«device.name.toFirstUpper»(ServiceConfiguration serviceConfiguration«attributeList») {
			    super("«device.className»", serviceConfiguration);
«««			    updateId(idValue);
			    «FOR attr : attributes»
			    	update«attr.name.toFirstUpper»(«attr.name»Value);
			    «ENDFOR»
				addToOutProperties("attributes", getAttributesCalled(null, null));
			}
		'''
	}
		   
    /* Template for functions inherated from the Service class */
    /**
     * 
     */
	def generateServiceFunctions(Iterable<AttributeDef> attrs, Iterable<OrderDef> actions, Iterable<SourceDef> sources) 
	'''		
		@Override
		protected final void internalPostInitialize() {
		    postInitialize();
		}
		
		@Override
		public final java.util.Map<String, Object> getAttributesCalled(java.util.Map<String, Object> properties, RemoteServiceInfo source) {
		    java.util.Map<String, Object> attributes = new java.util.HashMap<String, Object>();
		    «FOR attr : attrs»
		    	attributes.put("«attr.name»", _«attr.name»);
		    «ENDFOR»
		    	attributes.put("id", _id);
		    return attributes;
		}	
		
		@SuppressWarnings("all")
		@Override
		public final Object invokeActionCalled(java.util.Map<String, Object> properties, RemoteServiceInfo source, String actionName, Object... args) throws Exception {
		    «FOR action : actions»
		    	if (actionName.equals("«action.name»")) {
		    	    «action.name»(«action.params.asString»);
		    	    return null;
		    	}
		    «ENDFOR»
		    throw new InvocationException("Undefined action method name " + actionName);
		}
		
		@Override
		public Object getValueCalled(java.util.Map<String, Object> properties, RemoteServiceInfo source, String valueCalled, Object... args) throws Exception {
		    «FOR attr : attrs»
		    	if (valueCalled.equals("«attr.name»")) {
		    	    return _«attr.name»;
		    	}
		    «ENDFOR»
		    	if (valueCalled.equals("_id")) {
		    		return _id;
		    	}
		    «FOR src : sources»
		    	if (valueCalled.equals("«src.name»")) {
		    	    return get«src.name.toFirstUpper»(«src.indices.asString»);
		    	}
		    «ENDFOR»
		    throw new InvocationException("Undefined value " + valueCalled);
		}		
	'''	
    
    /* Template for attribute functions */
    /**
     * 
     */
	def generateAttributeFields(AttributeDef attribute) 
	'''
		private «attribute.attrTypeRef.fullName» _«attribute.name»;
	'''
    
    /* Template for attribute functions */
    /**
     * 
     */
	def generateAttributeMethods(AttributeDef attribute) 
	'''
		protected void update«attribute.name.toFirstUpper»(«attribute.attrTypeRef.fullName» new«attribute.name.toFirstUpper»Value) {
		    if (_«attribute.name» != new«attribute.name.toFirstUpper»Value) {
		        _«attribute.name» = new«attribute.name.toFirstUpper»Value;
		        updateAttribute("«attribute.name»", new«attribute.name.toFirstUpper»Value);
		    }
		}
		
		public «attribute.attrTypeRef.fullName» get«attribute.name.toFirstUpper»() {
		    return _«attribute.name»;
		}		
	'''
	
	def generateSourceFields(SourceDef source) {
		val hasIndices = (source.indices != null && source.indices.size > 0);
		val type = source.dataTypeRef.fullName;
		val hashType = "java.util.HashMap<" + source.name.toFirstUpper + "Indices, " + type + ">";
		
	'''
		«IF !hasIndices»private «type» _«source.name»;«ELSE»private «hashType» _«source.name» = new «hashType»();«ENDIF»
	'''
		
	}
	
	/* Template for source functions */
	/**
	 * 
	 */
	def generateSourceMethods(SourceDef source) {
		val hasIndices = (source.indices != null && source.indices.size > 0);
		val type = source.dataTypeRef.fullName;
				
	'''
		protected void publish«source.name.toFirstUpper»(«type» new«source.name.toFirstUpper»Value«IF hasIndices»,
				«source.indices.map[dataTypeRef.fullName + " " + name].join(",\n")»«ENDIF») {
			«IF hasIndices»
				«source.name.toFirstUpper»Indices _indices_ = new «source.name.toFirstUpper»Indices(«source.indices.variableNames»);
				_«source.name».put(_indices_, new«source.name.toFirstUpper»Value);
			«ELSE»
				_«source.name» = new«source.name.toFirstUpper»Value;
			«ENDIF»
			getProcessor().publishValue(getOutProperties(), "«source.name»", new«source.name.toFirstUpper»Value«IF hasIndices», «source.indices.variableNames»«ENDIF»); 
		}
		
		protected «type» get«source.name.toFirstUpper»(«source.indices.parameters») throws Exception {
			return _«source.name»«IF hasIndices».get(new «source.name.toFirstUpper»Indices(«source.indices.variableNames»))«ENDIF»;
		}
		
		protected abstract List<DataDeliveryModel> get«source.name.toFirstUpper»DeliveryModels();
	'''
	}
	
	/* Template for action functions */
	/**
	 * 
	 */
	def generateActionMethods(OrderDef action) '''
		protected abstract void «action.name»(«action.params.parameters») throws Exception;
	'''
	
	/* Generate the container class for publishing a source value */
	/**
	 * 
	 */
	def generateDeviceSourceValue(SourceDef source, DeviceType device) {		
		
		valueClass(
			source.name, 
			source.name.toFirstUpper + "From" + device.name,			
			device.packageNameForSource, 
			'''the source <code>«source.name»</code>''',
			device, 
			'''An object to store a value published on the source <code>«source.name»</code> of the device <code>«device.name»</code>.''',
			source.dataTypeRef, 
			source.indices,
			"this.sender = new " + device.name + "Proxy(service, remoteServiceInfo);\n",
			"Service service, RemoteServiceInfo remoteServiceInfo, ", 
			
			'''				
				// Proxy to the sender, i.e., the device that published the source
				private «device.name»Proxy sender;
				
				/**
				 * Get the sender of the source <code>«source.name»</code>, i.e., the <code>«device.name»</code> device that published the source.
				 * 
				 * @return A proxy to the <code>«device.name»</code> that triggered the source
				 */
				public «device.name»Proxy sender() {
				    return sender;
				}
				
				/**
				 * A proxy to the <code>«device.name»</code> device enclosing the subscription/unsubscription methods.
				 */
				public class «device.name»Proxy extends Proxy {
				    private «device.name»Proxy(Service service, RemoteServiceInfo remoteServiceInfo) {
				        super(service, remoteServiceInfo);
				    }
				    
				    /**
				     * Subcribes to publication of the <code>«source.name»</code> source from the <code>«device.name»</code> device.
				     */
				    public void subscribeTo«source.name.toFirstUpper»() {
				        getService().getProcessor().subscribeValue(getService().getOutProperties(), new fr.inria.diagen.core.service.filter.SubscriptionFilter(this.getRemoteServiceInfo()), "«source.name»");
				    }
				    
				    /**
				     * Unsubcribes from publication of the <code>«source.name»</code> source from the <code>«device.name»</code> device.
				     */
				    public void unsubscribe«source.name.toFirstUpper»() {
				        getService().getProcessor().unsubscribeValue(getService().getOutProperties(), new fr.inria.diagen.core.service.filter.SubscriptionFilter(this.getRemoteServiceInfo()), "«source.name»");
				    }
				    «FOR attr : device.allAttributes»
				    	
«««				    	/**
«««				    	 «IF !attr.comments.nullOrEmpty»
«««				    	 	* «attr.comments.split("\n").join("* ")»
«««				    	 	* 
«««				    	 «ENDIF»
«««				    	 * @return the value of the <code>«attr.name»</code> attribute
«««				    	 */
				    	public «attr.attrTypeRef.fullName» «attr.name»() {
				    	    return («attr.attrTypeRef.fullName») callGetValue("«attr.name»");
				    	}
				    «ENDFOR»
				}
			''', '''
				
				import fr.inria.diagen.core.network.RemoteServiceInfo;
				import fr.inria.diagen.core.service.local.Service;
				import fr.inria.diagen.core.service.proxy.Proxy;
			''', ""/*source.comments*/)
	}
	
	/* Generate the index class for an indexed source */
	def generateDeviceSourceIndices(SourceDef source, DeviceType device) {
		return classForIndices(
			source.indices,
			device.packageNameForSource,
			source.name,
			source,
			'''Container for the indices of the source <code>«source.name»</code> from device <code>«device.name»</code>.'''
		);
	}
	
/*============================================================================*/

}
