package fr.inria.phoenix.diasuite.diaswarm.generator.template.mock

import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DomainModel
import fr.inria.phoenix.diasuite.diaswarm.generator.template.DiaSwarmEntity

import static fr.inria.phoenix.diasuite.diaswarm.generator.template.mock.DiaSwarmMockHelper.*

import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.context.DiaSwarmContextHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.controller.DiaSwarmControllerHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.device.DiaSwarmDeviceHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.datatype.DiaSwarmDatatypeHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.deployment.DiaSwarmDeploymentHelper.*

class MockClassGen extends DiaSwarmEntity {
		
    def generateMockClass(Iterable<DomainModel> models) {
    	val packageName = mockPackageName
    	mockClass(models, packageName)
    }
    
    def String mockClass(Iterable<DomainModel> models, String packageNameVal) '''
        package «packageNameVal»;
        
        import fr.inria.diagen.core.ServiceConfiguration;
        import «packageName».AppComponentBinder;
        
        «FOR c : models.contexts»
        // Context import for «c.name»
        import «c.packageName».Abstract«c.name.toFirstUpper»;
        «ENDFOR»
        «FOR c : models.controllers»
        // Controller import for «c.name»
        import «c.packageName».Abstract«c.name.toFirstUpper»;
        «ENDFOR»
        
                
        /**
         * Use this class to test your DiaSuite application.
         *
         * <p>
         * Here is n example of usage to test an application that should trigger an action <code>action</code> on a device <code>TestDevice</code>
         * when it receive a source "src" from this device:
         * <pre>
         * ...
         * import static «packageNameVal».Mock.*;
         * ...
         *    @Before
         *    public void setUp() throws Exception {
         *        underTest(MyAppComponentBinder.class);
         *    }
         *    
         *    @After
         *    public void tearDown() {
         *        shutdown();
         *    }
         *    
         *    @Test
         *    public void testMyApplication() {
         *        assertTrue(mockTestDevice().src("hehe").expectAction());
         *    }
         * ...
         */
        @SuppressWarnings("all")
        public final class Mock extends AppComponentBinder {
            
            /**
             * The time out, in milliseconds, before considering an action will never happens.
             * It should be changed if the application has long computations.
             */
            public static long TIMEOUT = 1000;
            
            private static Mock mocker;
            
            /**
             * Put a class under test
             * @param binder The application class
             * @return The mocking factory
             */
            public static Mock underTest(Class<? extends AppComponentBinder> binder) throws Exception {
                mocker = new Mock(binder);
                return mocker;
            }
            
            /**
             * Tear-down the device mocking framework
             */
            public static void shutdown() {
                if(mocker != null)
                    mocker.undeployAll();
                mocker = null;
            }
            
            private AppComponentBinder delegate;
            private Mock(Class<? extends AppComponentBinder> binder) throws Exception {
                delegate = binder.newInstance();
                // Starts the framework
                deployAll();
            }
            
            @SuppressWarnings("unchecked")
            public ServiceConfiguration getServiceConfiguration(String name) {
                try {
                    Class<? extends ServiceConfiguration> clazz = (Class<? extends ServiceConfiguration>) Class.forName("fr.inria.diagen.commlayer.local.LocalServiceConfiguration");
                    java.lang.reflect.Constructor<? extends ServiceConfiguration> constructor = clazz.getConstructor(String.class);
                    return constructor.newInstance(name);
                } catch (Exception ex) {
                    fr.inria.diagen.log.DiaLog.critical("Unabled to initialize local communication layer. Please make sure correct libraries are in the classpath", ex);
                    System.exit(-1);
                    return null;
                }
            }
            
            private String randomId() {
                StringBuffer result = new StringBuffer();
                for (int i = 0; i < 10; i++)
                    result.append(Integer.toString((int) (Math.random() * 10)));
                return result.toString();
            }
            
        
          	// Delegation part
           	«FOR c : models.controllers»
           	public Class<? extends Abstract«c.name.toFirstUpper»> get«c.name.toFirstUpper»Class() {
           	return delegate.get«c.name.toFirstUpper»Class();
           	}
           	«ENDFOR»
           	«FOR c : models.contexts»
           	public Class<? extends Abstract«c.name.toFirstUpper»> get«c.name.toFirstUpper»Class() {
           	return delegate.get«c.name.toFirstUpper»Class();
           	}
           	«ENDFOR»
           	
           	// Mocks constructor for all devices
            	«FOR device : models.devices»
«««                «device.javadoc("\nReturns a mock of the device <code>" + device.name + "</code>.",
«««                    device.allAttributes.map["\n@param " + name + " the value of the <code>" + name + "</code> attribute"].join("") +
«««                    "\n@return a mock of the device <code>" + device.name + "</code>."
«««                )»
                	public «device.name»Mock «device.name.toFirstLower»(«device.allAttributes.map[attrTypeRef.fullName + " " + name].join(",\n        ")») {
                    	«device.name»Mock mock = new «device.name»Mock(getServiceConfiguration("«device.name»" + randomId())«device.allAttributes.map[", " + name].join("")»);
                    	deploy(mock.proxy);
                    	return mock;
                	}
                
«««                «device.javadoc("\nReturns a mock of the device <code>" + device.name + "</code>.",
«««                    device.allAttributes.map["\n@param " + name + " the value of the <code>" + name + "</code> attribute"].join("") +
«««                    "\n@return a mock of the device <code>" + device.name + "</code>."
«««                )»
                	public static «device.name»Mock mock«device.name»(«device.allAttributes.map[attrTypeRef.fullName + " " + name].join(",\n        ")») {
                    	if(mocker == null)
                        	throw new RuntimeException("Mock is not initialized");
                    	return mocker.«device.name.toFirstLower»(«device.allAttributes.map[name].join(", ")»);
                	}
            	«ENDFOR»
        }
    '''
}
