package fr.inria.phoenix.diasuite.diaswarm.generator.template.mock

import fr.inria.phoenix.diasuite.diaswarm.diaswarm.AttributeDef

import fr.inria.phoenix.diasuite.diaswarm.diaswarm.DeviceType
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.OrderDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.SourceDef
import fr.inria.phoenix.diasuite.diaswarm.diaswarm.VariableDef
import fr.inria.phoenix.diasuite.diaswarm.generator.template.DiaSwarmEntity
import org.eclipse.emf.ecore.EObject

import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.mock.DiaSwarmMockHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.datatype.DiaSwarmDatatypeHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.device.DiaSwarmDeviceHelper.*
import static extension fr.inria.phoenix.diasuite.diaswarm.generator.template.Utils.*

class MockDeviceGen extends DiaSwarmEntity {
	
	def generateMock(DeviceType device) {
		val String packageName = device.mockPackageName
		mockClassForDevice(device, packageName)
	}	
    
//    override visitDeviceDef(DeviceDef device) {
//        val String packageName = DiaSpecFramework::basePackage + ".mocks";
//        generateJavaFile(device.name + "Mock", packageName, mockClassForDevice(device, packageName));
//    }
    
    def String deviceName(EObject obj) {
        (obj.eContainer as DeviceType).name
    }
    
    /* Template for the constructor of the mock */
    def constructor(DeviceType device) {
        return '''
        '''
    }
    
    /* Template for attribute functions */
    // Proxy to bypass protection
    def proxyTemplate(AttributeDef attribute) '''
        public void _update«attribute.name.toFirstUpper»(«attribute.attrTypeRef.fullName» new«attribute.name.toFirstUpper»Value) {
            update«attribute.name.toFirstUpper»(new«attribute.name.toFirstUpper»Value);
        }
    '''
    
    def template(AttributeDef attribute, String className) '''
        // Code for the attribute «attribute.name» from device «attribute.deviceName»
«««        «attribute.javadoc(
«««            "\nSet the value of the <code>" + attribute.name + "</code> attribute from device <code>" + attribute.deviceName + "</code>.",
«««            "\n@param new" + attribute.name.toFirstUpper + "Value the new value of <code>" + attribute.name + "</code>" +
«««            "\n@return this for fluent interface",
«««            1
«««        )»
        public «className» «attribute.name.toFirstLower»(«attribute.attrTypeRef.fullName» new«attribute.name.toFirstUpper»Value) {
            proxy._update«attribute.name.toFirstUpper»(new«attribute.name.toFirstUpper»Value);
            return this;
        }
        // End of code for the attribute «attribute.name» from device «attribute.deviceName»
    '''
    
    /* Template for source functions */
    def template(SourceDef source, String className) {
        val hasIndices = (source.indices != null && source.indices.size > 0);
        val type = source.dataTypeRef.fullName;
        val hashType = "java.util.HashMap<" + source.name.toFirstUpper + "Indices, " + type + ">";
        '''
        // Code for source «source.name» from device «source.deviceName»
«««        «source.javadoc(
«««            "\nPublish the value of source <code>" + source.name + "</code> from device <code>" + source.deviceName + "</code>.",
«««            "\n@param new" + source.name.toFirstUpper + "Value the new value for the source <code>" + source.name + "</code>" +
«««            source.indices.map["\n" + '''@param «name» the value of the index <code>«name»</code>'''].join("") +
«««            "\n@return this for fluent interface",
«««            1
«««        )»
        public «className» «source.name.toFirstLower»(«type» new«source.name.toFirstUpper»Value«IF hasIndices»,
                «source.indices.map[dataTypeRef.fullName + " " + name].join(",\n")»«ENDIF») {
            proxy._publish«source.name.toFirstUpper»(new«source.name.toFirstUpper»Value«IF hasIndices»,
                «source.indices.map[name].join(",\n")»«ENDIF»);
            return this;
        }
        
        «IF !hasIndices»private «type» __«source.name» = null;«ELSE»private «hashType» __«source.name» = new «hashType»();«ENDIF»
«««        «source.javadoc(
«««            "\nSet the value (without publication) of source <code>" + source.name + "</code> from device <code>" + source.deviceName + "</code>.",
«««            "\n@param new" + source.name.toFirstUpper + "Value the new value for the source <code>" + source.name + "</code>" +
«««            source.indices.map["\n" + '''@param «name» the value of the index <code>«name»</code>'''].join("") +
«««            "\n@return this for fluent interface",
«««            1
«««        )»
        public «className» set«source.name.toFirstUpper»(«type» new«source.name.toFirstUpper»Value«IF hasIndices»,
                «source.indices.map[dataTypeRef.fullName + " " + name].join(",\n")»«ENDIF») {
            «IF hasIndices»
            «source.name.toFirstUpper»Indices _indices_ = new «source.name.toFirstUpper»Indices(«source.indices.variableNames»);
            __«source.name».put(_indices_, new«source.name.toFirstUpper»Value);
            «ELSE»
            __«source.name» = new«source.name.toFirstUpper»Value;
            «ENDIF»
            return this;
        }
        // End of code for source «source.name» from device «source.deviceName»
        '''
    }
    
    def proxyTemplate(SourceDef source) {
        val hasIndices = (source.indices != null && source.indices.size > 0);
        val type = source.dataTypeRef.fullName;
        '''
        protected «type» get«source.name.toFirstUpper»(«source.indices.parameters») throws Exception {
            «type» result = __«source.name»«IF hasIndices».get(new «source.name.toFirstUpper»Indices(«source.indices.variableNames»))«ENDIF»;
            if(result == null)
                return super.get«source.name.toFirstUpper»(«source.indices.map[name].join(", ")»);
            return result;
        }

        public void _publish«source.name.toFirstUpper»(«type» new«source.name.toFirstUpper»Value«IF hasIndices»,
                «source.indices.map[dataTypeRef.fullName + " " + name].join(",\n")»«ENDIF») {
            publish«source.name.toFirstUpper»(new«source.name.toFirstUpper»Value«IF hasIndices»,
                «source.indices.map[name].join(",\n")»«ENDIF»);
        }
        '''
    }
    
    /* Template for action functions */
    def proxyTemplate(OrderDef order) '''
        protected void «order.name»(«order.params.parameters») throws Exception {
            synchronized(__actions) {
                java.util.Queue<Object> __action = new java.util.LinkedList<Object>();
                __action.add("«order.name»");
                «FOR x : order.params»
                    __action.add(«x.name»);
                «ENDFOR»
                __actions.add(__action);
                __actions.notifyAll();
             }
        }
    '''
        
    def template(OrderDef order) '''
        «FOR y : order.params.index»
«««            «order.javadoc('''
«««            Check that the <code>«order.name»</code> order from the <code>«action.action.name»</code> action
«««            defined in device «action.deviceName» was called.
«««            ''', '''
«««            «FOR x: order.params.index.filter[key < y.key]»
«««                
«««                @param «x.value.name» parameter «x.key+1» of the order.«IF !Utils::getComments(x.value, 1).nullOrEmpty» «Utils::getComments(x.value, 1)»«ENDIF»«ENDFOR»
«««                @return true if the action happened with the given parameters, remaining parameters are ignored
«««            ''', 1)»
            «expectMethod(order.name, order.params.index.filter[key < y.key].map[value])»
        «ENDFOR»
«««        «order.javadoc('''
«««            Check that the <code>«order.name»</code> order from the <code>«action.action.name»</code> action
«««            defined in device «action.deviceName» was called.
«««            ''', '''
«««            «FOR x: order.params.index»
«««                
«««                @param «x.value.name» parameter «x.key+1» of the order.«IF !Utils::getComments(x.value, 1).nullOrEmpty» «Utils::getComments(x.value, 1)»«ENDIF»«ENDFOR»
«««                @return true if the action happened with the given parameters
«««            ''', 1)»
        «expectMethod(order.name, order.params)»
    '''
    
    def expectMethod(String orderName, Iterable<VariableDef> params)
        '''
        public boolean expect«orderName.toFirstUpper»(«params.parameters») {
            try {
                synchronized(__actions) {
                    java.util.Queue<Object> __action = __actions.poll();
                    if(__action == null) {
                        __actions.wait(TIMEOUT);
                        __action = __actions.poll();
                    }
                    if(__action == null)
                        return false;
                    if(!("«orderName»".equals(__action.poll())))
                        return false;
                    «FOR x : params»
                        if(!«x.name».equals(__action.poll()))
                            return false;
                    «ENDFOR»
                    return true;
                }
             } catch(InterruptedException e) {
                 return false;
             }
        }
        '''
    
    /* Template for the mock class for the device */
    def mockClassForDevice(DeviceType device, String packageName) {
        val attrs = device.allAttributes;
        val srcs = device.allSources.sort([a, b | a.name.compareTo(b.name)]);
        val actions = device.allActionMethods
        val orders = actions.sort([a, b | a.name.compareTo(b.name)]);
        return '''
            package «packageName»;
            
            import fr.inria.diagen.core.ServiceConfiguration;
            import «device.packageName».Abstract«device.name.toFirstUpper»;
            «FOR src : device.allSources.filter[!indices.nullOrEmpty]»
                import «device.packageName».source.«src.name.toFirstUpper»Indices;
            «ENDFOR»
            
            // @internal
            public final class «device.name.toFirstUpper»Mock {
                
                /**
                 * The timeout in milliseconds to wait action. Set it if your application computations are long.
                 */
                public long TIMEOUT = Mock.TIMEOUT;
                
                class Proxy extends Abstract«device.name.toFirstUpper» {
                    private Proxy(ServiceConfiguration serviceConfiguration«attrs.map[",\n        " + attrTypeRef.fullName + " " + name + "Value"].join("")») {
                        super(serviceConfiguration«attrs.sortBy[attr | attr.name].map[", " + name + "Value"].join("")»);
                    }
                    «FOR attr: attrs»
                        
                        «attr.proxyTemplate»
                    «ENDFOR»
                    «FOR source : srcs»
                        
                        «source.proxyTemplate»
                    «ENDFOR»
                    «FOR act: orders»
                        
                        «act.proxyTemplate»
                    «ENDFOR»
                }
                
                Proxy proxy;
                
                «device.name.toFirstUpper»Mock(ServiceConfiguration serviceConfiguration«attrs.map[",\n        " + attrTypeRef.fullName + " " + name + "Value"].join("")») {
                    this.proxy = new Proxy(serviceConfiguration«attrs.map[", " + name + "Value"].join("")»);
                }
                
                «FOR attr: attrs»
                    
                    «attr.template(device.name.toFirstUpper + "Mock")»
                «ENDFOR»
                «FOR source: srcs»
                    
                    «source.template(device.name.toFirstUpper + "Mock")»
                «ENDFOR»
                «IF !orders.nullOrEmpty»
                    
                    private java.util.Queue<java.util.Queue<Object>> __actions = new java.util.LinkedList<java.util.Queue<Object>>();
                «ENDIF»
                «FOR act: orders»
                    
                    «act.template()»
                «ENDFOR»
            }
    ''' }

}
